(*
================================================================================
dFlickr, API Kit for Flickr
Version: 0.8, 2005-03-06
================================================================================
Author : Dimitris Giannitsaros
Email  : daremon@hol.gr
URL    : http://www.daremon.gr/flickr/
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
License ('Do anything you want, don't blame me for anything' style):

Copyright (c) 2005, Dimitris Giannitsaros

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
================================================================================
*)

// TODO: ! UTF-8 support
// TODO: ! Default function parameters
// TODO: ! Implement upload

unit dFlickr;

interface

uses
  Contnrs, Classes, IdBaseComponent, IdComponent, IdTCPConnection, IdTCPClient,
  IdHTTP, SysUtils, LibXMLParser;

const
  FLICKR_NULL_NUMBER = -2;
  FLICKR_NULL_STRING = #1#2#3#4#5;
  FLICKR_BASE_URL_METHOD = 'http://flickr.com/services/rest/?method=';

  FlickrSortLookup:  array[0..3] of string = ('date-posted-asc', 'date-posted-desc',
                                              'date-taken-asc', 'date-taken-desc');
  FlickSortDatePostedAsc  = 0;
  FlickSortDatePostedDesc = 1;
  FlickSortDateTakenAsc   = 2;
  FlickSortDateTakenDesc  = 3;

type

  TFlickrParam = record
    Key: string;
    Value: string;
  end;
  TFlickrParamArray = array of TFlickrParam;

  EFlickrError = class(Exception);

  TFlickrError = record
    Xml: string;
    ErrorNo: string;
    ErrorDesc: string;
  end;

  TFlickr = class;
  TFlickrPeopleList = class;
  TFlickrPhotosetList = class;

  // Inherited by all classes that have a correspondent response XML
  // These classes must implement Parse(), in order to fill their appropriate properties
  TFlickrXml = class
  private
    FXml: string;
    FXMLParser: TXmlParser;
    procedure ParseInitialize(Xml: string);
    procedure Parse; virtual;
    procedure ParseFinalize;
    function ParseError(Xml: string): TFlickrError;
  protected
    procedure SetAttr(AttrName: string; var Variable: string); overload;
    procedure SetAttr(AttrName: string; var Variable: Integer); overload;
  public
    procedure GenerateFromXML(Xml: string; var Error: TFlickrError);
    function ParseNotesAdd(Xml: string): string;
    function ParseUrlsGetGroup(Xml: string): string;
    function ParseUrlsGetUserUrl(Xml: string): string;
    function ParseUrlsLookupGroup(Xml: string): string;
    function ParseUrlsLookupUser(Xml: string): string;
  end;

  // Note class
  TFlickrNote = class
  private
    FId: string;
    FAuthorId: string;
    FAuthorname: string;
    FX, FY, FW, FH: Integer;
    FText: string;
  protected
    FFlickr: TFlickr; // Reference to TFlickr creator
  public
    property Id: string read FId;
    property AuthorId: string read FAuthorId;
    property Authorname: string read FAuthorname;
    property X: Integer read FX;
    property Y: Integer read FY;
    property W: Integer read FW;
    property H: Integer read FH;
    property Text: string read FText;
    constructor Create(Flickr: TFlickr);
  end;

  // Notelist class
  TFlickrNoteList = class
  private
    FNoteList: TObjectList;
    function AddNote(PhotoId: string; X, Y, W, H: Integer; Text: string): TFlickrNote;
    function EdtNote(NoteId: string; X, Y, W, H: Integer; Text: string): Boolean;
    function DelNote(NoteId: string): Boolean;
  protected
    FFlickr: TFlickr; // Reference to TFlickr creator
  public
    constructor Create(Flickr: TFlickr);
    destructor Destroy; override;
    function Count: Integer;
    function GetNote(i: Integer): TFlickrNote;
  end;

  // Tag class
  TFlickrTag = class
  private
    FId: string;
    FAuthorId: string;
    FText: string;
    FCount: Integer;  // Only filled when TagsPopular is used
  protected
    FFlickr: TFlickr; // Reference to TFlickr creator
  public
    property Id: string read FId;
    property AuthorId: string read FAuthorId;
    property Text: string read FText;
    property Count: Integer read FCount;
    constructor Create(Flickr: TFlickr);
  end;

  // TagList class
  TFlickrTagList = class(TFlickrXml)
  private
    FTagList: TObjectList;
    procedure Parse; override;
  protected
    FFlickr: TFlickr; // Reference to TFlickr creator
  public
    constructor Create(Flickr: TFlickr);
    destructor Destroy; override;
    function Count: Integer;
    function GetTag(i: Integer): TFlickrTag;
  end;

  // Size class
  TFlickrSize = class
  private
    FText: string;
    FSource: string;
    FUrl: string;
    FWidth: Integer;
    FHeight: Integer;
  protected
    FFlickr: TFlickr; // Reference to TFlickr creator
  public
    property Text: string read FText;
    property Source: string read FSource;
    property Url: string read FUrl;
    property Width: Integer read FWidth;
    property Height: Integer read FHeight;
    constructor Create(Flickr: TFlickr);
  end;

  // Sizelist class
  TFlickrSizeList = class(TFlickrXml)
  private
    FSizeList: TObjectList;
    procedure Parse; override;
  protected
    FFlickr: TFlickr; // Reference to TFlickr creator
  public
    constructor Create(Flickr: TFlickr);
    destructor Destroy; override;
    function Count: Integer;
    function GetSize(i: Integer): TFlickrSize;
  end;

  // Photo class
  TFlickrPhoto = class(TFlickrXml)
  private
    FId: string;
    FOwnerId: string;
    FOwnerUsername: string;
    FOwnerRealname: string;
    FOwnerLocation: string;
    FSecret: string;
    FServer: string;
    FTitle: string;
    FDescription: string;
    FIsPublic: string;
    FIsFriend: string;
    FIsFamily: string;
    FIsFavorite: string;
    FLicense: string;
    FDatePosted: string;
    FDateTaken: string;
    FDateTakenGranularity: string;
    FPermComment: string;
    FPermAddMeta: string;
    FCanComment: string;
    FCanAddMeta: string;
    FCommentsCount: Integer;
    FIsPrimary: string; // only filled when list is created by Photoset.Photos
    FNotes: TFlickrNoteList;
    FTags: TFlickrTagList;
    FSizes: TFlickrSizeList;
    procedure Parse; override;
    // The first time these functions are called they retrieve and cache the
    // results from Flickr. The cache is NEVER invalidated.
    function GetOwnerId: string;
    function GetOwnerUsername: string;
    function GetOwnerRealname: string;
    function GetOwnerLocation: string;
    function GetSecret: string;
    function GetServer: string;
    function GetTitle: string;
    function GetDescription: string;
    function GetIsPublic: Boolean;
    function GetIsFriend: Boolean;
    function GetIsFamily: Boolean;
    function GetIsFavorite: Boolean;
    function GetLicense: string;
    function GetDatePosted: string;
    function GetDateTaken: string;
    function GetDateTakenGranularity: string;
    function GetPermComment: string;
    function GetPermAddMeta: string;
    function GetCanComment: string;
    function GetCanAddMeta: string;
    function GetCommentsCount: Integer;
    function GetNotes: TFlickrNoteList;
    function GetTags: TFlickrTagList;
    function GetSizes: TFlickrSizeList;
  protected
    FFlickr: TFlickr; // Reference to TFlickr creator
  public
    property Id: string read FId;
    property OwnerId: string read GetOwnerId;
    property OwnerUsername: string read GetOwnerUsername;
    property OwnerRealname: string read GetOwnerRealname;
    property OwnerLocation: string read GetOwnerLocation;
    property Secret: string read GetSecret;
    property Server: string read GetServer;
    property Title: string read GetTitle;
    property Description: string read GetDescription;
    property IsPublic: Boolean read GetIsPublic;
    property IsFriend: Boolean read GetIsFriend;
    property IsFamily: Boolean read GetIsFamily;
    property IsFavorite: Boolean read GetIsFavorite;
    property License: string read GetLicense;
    property DatePosted: string read GetDatePosted;
    property DateTaken: string read GetDateTaken;
    property DateTakenGranularity: string read GetDateTakenGranularity;
    property PermComment: string read GetPermComment;
    property PermAddMeta: string read GetPermAddMeta;
    property CanComment: string read GetCanComment;
    property CanAddMeta: string read GetCanAddMeta;
    property CommentsCount: Integer read GetCommentsCount;
    property Tags: TFlickrTagList read GetTags;
    property Sizes: TFlickrSizeList read GetSizes;
    property Notes: TFlickrNoteList read GetNotes;
    procedure ResetProperties;
    function NotesAdd(X, Y, W, H: Integer; Text: string): TFlickrNote;
    function NotesEdit(NoteId: string; X, Y, W, H: Integer; Text: string): Boolean;
    function NotesDelete(NoteId: string): Boolean;
    function FavoritesAdd: Boolean;
    function FavoritesRemove: Boolean;
    function TagsAdd(PTags: string): Boolean;
    function SetDates(PDatePosted: string = '';
                      PDateTaken: string = ''; PDateTakenGranularity: string = ''): Boolean;
    function SetMeta(PTitle, PDescription: string): Boolean;
    function SetPerms(PIsPublic, PIsFriend, PIsFamily: Boolean;
                      PPermComment, PPermAddMeta: Integer): Boolean;
    function SetTags(PTags: string): Boolean;
    constructor Create(Flickr: TFlickr);
    destructor Destroy; override;
  end;

  // PhotoList class
  TFlickrPhotoList = class(TFlickrXml)
  private
    FPage: Integer;
    FPages: Integer;
    FPerPage: Integer;
    FTotal: Integer;
    FPhotoList: TObjectList;
    procedure Parse; override;
  protected
    FFlickr: TFlickr; // Reference to TFlickr creator
  public
    property Page: Integer read FPage;
    property Pages: Integer read FPages;
    property PerPage: Integer read FPerPage;
    property Total: Integer read FTotal;
    constructor Create(Flickr: TFlickr);
    destructor Destroy; override;
    function Count: Integer;
    function GetPhoto(i: Integer): TFlickrPhoto;
  end;

  // People class
  TFlickrPeople = class(TFlickrXml)
  private
    FId: string;
    FIsAdmin: string;
    FIsPro: string;
    FUsername: string;
    FRealname: string;
    FLocation: string;
    FPhotosFirstDate: string;
    FPhotosFirstDateTaken: string;
    FPhotosCount: Integer;
    FUrlPhotos: string;
    FUrlProfile: string;
    // Next 2 fields are only filled from a call to Flickr.PeopleGetOnlineList
    FOnline: string;       // 2 means fully online, 1 means away
    FAwayMessage: string;  // away message
    FPublicPhotos: TFlickrPhotoList;
    FContactsPublicPhotos: TFlickrPhotoList;
    FPublicContacts: TFlickrPeopleList;
    FTags: TFlickrTagList;
    FTagsPopular: TFlickrTagList;
    FFavorites: TFlickrPhotoList;
    FFavoritesPublic: TFlickrPhotoList;
    FPhotosets: TFlickrPhotosetList;
    procedure Parse; override;
    // The first time these functions are called they retrieve and cache the
    // results from Flickr. The cache is NEVER invalidated.
    function GetIsAdmin: Boolean;
    function GetIsPro: Boolean;
    function GetUsername: string;
    function GetRealname: string;
    function GetLocation: string;
    function GetPhotosFirstDate: string;
    function GetPhotosFirstDateTaken: string;
    function GetPhotosCount: Integer;
    function GetPublicPhotos: TFlickrPhotoList;
    function GetPublicContacts: TFlickrPeopleList;
    function GetTags: TFlickrTagList;
    function GetTagsPopular: TFlickrTagList;
    function GetFavorites: TFlickrPhotoList;
    function GetFavoritesPublic: TFlickrPhotoList;
    function GetPhotosets: TFlickrPhotosetList;
    function GetUrlPhotos: string;
    function GetUrlProfile: string;
  protected
    FFlickr: TFlickr; // Reference to TFlickr creator
  public
    property Id: string read FId;
    property IsAdmin: Boolean read GetIsAdmin;
    property IsPro: Boolean read GetIsPro;
    property Username: string read GetUsername;
    property Realname: string read GetRealname;
    property Location: string read GetLocation;
    property PhotosFirstDate: string read GetPhotosFirstDate;
    property PhotosFirstDateTaken: string read GetPhotosFirstDateTaken;
    property PhotosCount: Integer read GetPhotosCount;
    property Online: string read FOnline;
    property AwayMessage: string read FAwayMessage;
    property UrlPhotos: string read GetUrlPhotos;
    property UrlProfile: string read GetUrlProfile;
    property PublicPhotos: TFlickrPhotoList read GetPublicPhotos;
    property PublicContacts: TFlickrPeopleList read GetPublicContacts;
    property Tags: TFlickrTagList read GetTags;
    property TagsPopular: TFlickrTagList read GetTagsPopular;
    property Favorites: TFlickrPhotoList read GetFavorites;
    property FavoritesPublic: TFlickrPhotoList read GetFavoritesPublic;
    property Photosets: TFlickrPhotosetList read GetPhotosets;
    constructor Create(Flickr: TFlickr);
    destructor Destroy; override;
    procedure ResetProperties;
    function GetContactsPublicPhotos(PJustFriends, PSinglePhoto,
                                     PIncludeSelf: Boolean; PCount: Integer = -1): TFlickrPhotoList;
  end;

  // PeopleList class
  TFlickrPeopleList = class(TFlickrXml)
  private
    FPeopleList: TObjectList;
    procedure Parse; override;
  protected
    FFlickr: TFlickr; // Reference to TFlickr creator
  public
    constructor Create(Flickr: TFlickr);
    destructor Destroy; override;
    function Count: Integer;
    function GetPeople(i: Integer): TFlickrPeople;
  end;

  // Photoset class
  TFlickrPhotoset = class(TFlickrXml)
  private
    FId: string;
    FTitle: string;
    FDescription: string;
    FUrl: string;
    FPrimaryPhotoId: string;
    FOwnerId: string;
    FPhotos: TFlickrPhotoList;
    procedure Parse; override;
    function GetOwnerId: string;
    function GetPrimaryPhotoId: string;
    function GetTitle: string;
    function GetDescription: string;
    function GetPhotos: TFlickrPhotoList;
  protected
    FFlickr: TFlickr; // Reference to TFlickr creator
  public
    property Id: string read FId;
    property Title: string read GetTitle;
    property Description: string read GetDescription;
    property Url: string read FUrl;
    property PrimaryPhoto: string read GetPrimaryPhotoId;
    property OwnerId: string read GetOwnerId;
    property Photos: TFlickrPhotoList read GetPhotos;
    function EditMeta(Title, Description: string): Boolean;
    function EditPhotos(PhotoIds: string): Boolean;
    constructor Create(Flickr: TFlickr);
    destructor Destroy; override;
    procedure ResetProperties;
  end;

  // PhotosetList class
  TFlickrPhotosetList = class(TFlickrXml)
  private
    FPhotosetList: TObjectList;
    procedure Parse; override;
  protected
    FFlickr: TFlickr; // Reference to TFlickr creator
  public
    constructor Create(Flickr: TFlickr);
    destructor Destroy; override;
    function Count: Integer;
    function GetPhotoset(i: Integer): TFlickrPhotoset;
  end;

  // Group class
  TFlickrGroup = class(TFlickrXml)
  private
    FId: string;
    FName: string;
    FMembers: Integer;
    FOnline: Integer;
    FPrivacy: string;
    FChatId: string;
    FChatCount: Integer;
    FUrl: string;
    // next 2, are only filled sometimes
    FAdmin: string;
    FPhotosCount: Integer;
    FPhotos: TFlickrPhotoList;
    procedure Parse; override;
    function GetName: string;
    function GetMembers: Integer;
    function GetOnline: Integer;
    function GetPrivacy: string;
    function GetChatId: string;
    function GetChatCount: Integer;
    function GetUrl: string;
  protected
    FFlickr: TFlickr; // Reference to TFlickr creator
  public
    property Id: string read FId;
    property Name: string read GetName;
    property Members: Integer read GetMembers;
    property Online: Integer read GetOnline;
    property Privacy: string read GetPrivacy;
    property ChatId: string read GetChatId;
    property ChatCount: Integer read GetChatCount;
    property Admin: string read FAdmin;
    property PhotosCount: Integer read FPhotosCount;
    property Url: string read GetUrl;
    constructor Create(Flickr: TFlickr);
    destructor Destroy; override;
    procedure ResetProperties;
    function PoolPhotoAdd(Photo: TFlickrPhoto): Boolean;
    function PoolPhotoRemove(Photo: TFlickrPhoto): Boolean;
    function GetPhotos(PTags: string = ''; PPerPage: Integer = -1;
                       PPage: Integer = -1): TFlickrPhotoList;
  end;

  // Group class (returned by flickr.groups.browse)
  TFlickrGroupCat = class
  private
    FId: string;
    FName: string;
    FGroupCount: Integer; // number of groups in Category
  public
    property Id: string read FId;
    property Name: string read FName;
    property GroupCount: Integer read FGroupCount;
  end;

  // GroupList class
  TFlickrGroupList = class(TFlickrXml)
  private
    FGroupList: TObjectList;
    FGroupCatList: TObjectList;
    procedure Parse; override;
  protected
    FFlickr: TFlickr; // Reference to TFlickr creator
  public
    constructor Create(Flickr: TFlickr);
    destructor Destroy; override;
    function GroupCount: Integer;
    function GetGroup(i: Integer): TFlickrGroup;
    function GroupCatCount: Integer;
    function GetGroupCat(i: Integer): TFlickrGroupCat;
  end;

  // Blog class
  TFlickrBlog = class
  private
    FId: string;
    FName: string;
    FUrl: string;
    FNeedPassword: string;
    function GetNeedPassword: Boolean;
  protected
    FFlickr: TFlickr; // Reference to TFlickr creator
  public
    property Id: string read FId;
    property Name: string read FName;
    property Url: string read FUrl;
    property NeedPassword: Boolean read GetNeedPassword;
    constructor Create(Flickr: TFlickr);
    function PostPhoto(Photo: TFlickrPhoto; Title, Description: string;
                       BlogPassword: string =''): Boolean;
  end;

  // Bloglist class
  TFlickrBloglist = class(TFlickrXml)
  private
    FBlogList: TObjectList;
    procedure Parse; override;
  protected
    FFlickr: TFlickr; // Reference to TFlickr creator
  public
    constructor Create(Flickr: TFlickr);
    destructor Destroy; override;
    function Count: Integer;
    function GetBlog(i: Integer): TFlickrBlog;
  end;

  // Main class. Inherits from TFlickrXml so it can parse plain responses (OK/Failed)
  // using GenerateFromXml
  // Communicates with Flickr.com through DoRequest()
  // All Flickr functions are included here
  TFlickr = class(TFlickrXml)
  private
    FEmail: string;
    FPassword: string;
    FAPIKey: string;
    FHTTP: TIdHTTP;
    function DoRequest(Method: string; Params: TFlickrParamArray): string;
    procedure Parse; override;
  protected
    // Photos section
    function PhotosGetInfo(Photo: TFlickrPhoto): Boolean;
    function PeopleGetPublicPhotos(UserId: string; PerPage, Page: Integer): TFlickrPhotoList;
    function PhotosGetSizes(Photo: TFlickrPhoto): Boolean;
    function PhotosAddTags(PhotoId, Tags: string): Boolean;
    function PhotosGetContext(PhotoId: string; var PrevPhoto,
                              NextPhoto: TFlickrPhoto): Boolean; // not implemented
    function PhotosSetDates(PhotoId: string; DatePosted: string = '';
                            DateTaken: string = ''; DateTakenGranularity: string = ''): Boolean;
    function PhotosGetExif(PhotoId, Secret: string): Boolean; // not implemented
    function PhotosGetPerms(PhotoId: string): Boolean; // not implemented
    function PhotosSetMeta(PhotoId, Title, Description: string): Boolean;
    function PhotosSetPerms(PhotoId: string; IsPublic, IsFriend, IsFamily: Boolean;
                            PermComment, PermAddMeta: Integer): Boolean;
    function PhotosSetTags(PhotoId, Tags: string): Boolean;
    function PhotosGetContactsPublicPhotos(UserId: string; JustFriends, SinglePhoto,
                                           IncludeSelf: Boolean; Count: Integer = -1): TFlickrPhotoList;
    // People section
    function PeopleGetInfo(People: TFlickrPeople): Boolean;
    // Contacts section
    function ContactsGetPublicList(UserId: string): TFlickrPeopleList;
    // Notes section
    function NotesAdd(PhotoId: string; NoteX, NoteY, NoteW,
                      NoteH: Integer; NoteText: string): TFlickrNote;
    function NotesEdit(NoteId: string; NoteX, NoteY, NoteW,
                      NoteH: Integer; NoteText: string): Boolean;
    function NotesDelete(NoteId: string): Boolean;
    // Tags section
    function TagsGetListUser(UserId: string): TFlickrTagList;
    function TagsGetListUserPopular(UserId: string; Count: Integer): TFlickrTagList;
    function TagsGetListPhoto(PhotoId: string): Boolean; // not implemented, covered by TFlickrPhoto.Tags
    // Favorites section
    function FavoritesAdd(PhotoId: string): Boolean;
    function FavoritesGetList(UserId: string; PerPage, Page: Integer): TFlickrPhotoList;
    function FavoritesGetPublicList(UserId: string; PerPage, Page: Integer): TFlickrPhotoList;
    function FavoritesRemove(PhotoId: string): Boolean;
    // Photosets section
    function PhotosetsEditMeta(PhotosetId, Title, Description: string): Boolean;
    function PhotosetsGetInfo(Photoset: TFlickrPhotoset): Boolean;
    function PhotosetsGetList(UserId: string): TFlickrPhotosetList;
    function PhotosetsGetPhotos(PhotosetId: string): TFlickrPhotoList;
    function PhotosetsEditPhotos(PhotosetId, PrimaryPhotoId, PhotoIds: string): Boolean;
    function PhotosetsGetContext(PhotoId, PhotosetId: string): Boolean; // not implemented
    // Blogs section
    function BlogsPostPhoto(BlogId, PhotoId, Title, Description: string;
                            BlogPassword: string = ''): Boolean;
    // Groups section
    function GroupsGetInfo(Group: TFlickrGroup): Boolean;
    // GroupsPools setion
    function GroupsPoolsAdd(PhotoId, GroupId: string): Boolean;
    function GroupsPoolsRemove(PhotoId, GroupId: string): Boolean;
    function GroupsPoolsGetPhotos(GroupId: string; Tags: string = ''; PerPage: Integer = -1;
                                  Page: Integer = -1): TFlickrPhotoList;
    function GroupsPoolsGetContext(PhotoId, GroupId: string): Boolean; // not implemented
    // Urls setion
    function UrlsGetGroup(GroupId: string): string;
    function UrlsGetUserPhotos(UserId: string): string;
    function UrlsGetUserProfile(UserId: string): string;
  public
    LastError: TFlickrError; // filled when an exception is raised
    constructor Create(APIKey: string);
    destructor Destroy; override;
    procedure Authenticate(Email, Password: string);

    // Uploading photos
    function Upload(FilePath: string): TFlickrPhoto;
    // Photos section
    function PhotosGetContactsPhotos(Count: Integer; JustFriends, SinglePhoto,
                                     IncludeSelf: Boolean): TFlickrPhotoList;
    function PhotosGetCounts(Dates, TakenDates: string): Boolean; // not implemented
    function PhotosGetNotInSet(PerPage, Page: Integer): TFlickrPhotoList;
    function PhotosGetRecent(PerPage, Page: Integer): TFlickrPhotoList;
    function PhotosGetUntagged(PerPage, Page: Integer): TFlickrPhotoList;
    function PhotosRemoveTag(TagId: string): Boolean;
    function PhotosSearch(UserId: string; Tags: string = ''; TagMode: string = ''; Text: string = '';
                          MinUploadDate: string = ''; MaxUpLoadDate: string = '';
                          MinTakenDate: string = ''; MaxTakenDate: string = ''; License: string = '';
                          PerPage: Integer = -1; Page: Integer = -1; Sort: Integer = -1): TFlickrPhotoList;
    // People section
    function PeopleFindByEmail(FindEmail: string): TFlickrPeople;
    function PeopleFindByUsername(Username: string): TFlickrPeople;
    function PeopleGetOnlineList: TFlickrPeopleList;
    // Contacts section
    function ContactsGetList: TFlickrPeopleList;
    // Photosets section
    function PhotosetsCreate(Title, Description, PrimaryPhotoId: string): TFlickrPhotoSet;
    function PhotosetsDelete(PhotosetId: string): Boolean;
    function PhotosetsOrderSets(PhotosetIds: string): Boolean;
    // Blogs section
    function BlogsGetList: TFlickrBlogList;
    // Groups
    function GroupsGetActiveList: TFlickrGroupList;
    function GroupsBrowse(CatId: string = ''): TFlickrGroupList;
    // GroupPools section
    function GroupsPoolsGetGroups: TFlickrGroupList;
    // Urls section
    function UrlsLookupGroup(Url: string): string;
    function UrlsLookupUser(Url: string): string;
  end;

implementation

{ TFlickrPhotoList }

function TFlickrPhotoList.Count: Integer;
begin
  Result := FPhotoList.Count;
end;

constructor TFlickrPhotoList.Create(Flickr: TFlickr);
begin
  FFlickr := Flickr;
  FPhotoList := TObjectList.Create;
end;

destructor TFlickrPhotoList.Destroy;
begin
  FPhotoList.Free;
  inherited;
end;

function TFlickrPhotoList.GetPhoto(i: Integer): TFlickrPhoto;
begin
  Result := TFlickrPhoto(FPhotoList[i]);
end;

procedure TFlickrPhotoList.Parse;
var
  Photo: TFlickrPhoto;
begin
  while FXmlParser.Scan do
  begin
    case FXmlParser.CurPartType of
      ptStartTag,
      ptEmptyTag:
        begin
          // read attributes
          if (FXmlParser.CurName = 'photos') then
          begin
            if (FXmlParser.CurAttr.Count > 0) then
            begin
              SetAttr('page', FPage);
              SetAttr('pages', FPages);
              SetAttr('perpage', FPerPage);
              SetAttr('total', FTotal);
            end;
          end;
          if (FXmlParser.CurName = 'photo') then
          begin
            if (FXmlParser.CurAttr.Count > 0) then
            begin
              Photo := TFlickrPhoto.Create(FFlickr);
              SetAttr('id', Photo.FId);
              SetAttr('owner', Photo.FOwnerId);
              SetAttr('secret', Photo.FSecret);
              SetAttr('server', Photo.FServer);
              SetAttr('title', Photo.FTitle);
              SetAttr('ispublic', Photo.FIsPublic);
              SetAttr('isfriend', Photo.FIsFriend);
              SetAttr('isfamily', Photo.FIsFamily);
              SetAttr('isprimary', Photo.FIsPrimary);
              FPhotoList.Add(Photo);
            end;
          end;
        end;
      ptContent,
      ptCData:
        begin
        end;
      else
        begin
        end;
    end;
  end;
end;

{ TFlickr }

constructor TFlickr.Create(APIKey: string);
begin
  FAPIKey := APIKey;
  FEmail := '';
  FPassword := '';
  FHTTP := TIdHTTP.Create(nil);
end;

destructor TFlickr.Destroy;
begin
  FHTTP.Free;
  inherited;
end;

procedure TFlickr.Authenticate(Email, Password: string);
begin
  FEmail := Email;
  FPassword := Password;
end;

function TFlickr.Upload(FilePath: string): TFlickrPhoto;
var
  Photo: TFlickrPhoto;
begin
  Photo := TFlickrPhoto.Create(Self);
  Result := Photo;
end;

// Do a request and return the XML response
function TFlickr.DoRequest(Method: string; Params: TFlickrParamArray): string;
var
  Url: string;
  i: Integer;
begin
  Url := FLICKR_BASE_URL_METHOD + Method + '&api_key=' + FAPIKey;
  // is user is authenticated add his credentials
  if (FEmail <> '') and (FPassword <> '') then
  begin
    Url := Url + '&email=' + FEmail;
    Url := Url + '&password=' + FPassword;
  end;
  // add rest params
  for i := Low(Params) to High(Params) do
    Url := Url + '&' + Params[i].Key + '=' + Params[i].Value;
  // return the response XML
  Result := FHTTP.Get(Url);
end;

function TFlickr.PhotosSearch(UserId: string; Tags: string = ''; TagMode: string = ''; Text: string = '';
                    MinUploadDate: string = ''; MaxUpLoadDate: string = '';
                    MinTakenDate: string = ''; MaxTakenDate: string = ''; License: string = '';
                    PerPage: Integer = -1; Page: Integer = -1; Sort: Integer = -1): TFlickrPhotoList;
var
  Params: TFlickrParamArray;
  Xml: string;
  PhotoList: TFlickrPhotoList;
begin
  SetLength(Params, 12);
  Params[0].Key := 'user_id';
  Params[0].Value := UserId;
  Params[1].Key := 'tags';
  Params[1].Value := Tags;
  Params[2].Key := 'tag_mode';
  Params[2].Value := TagMode;
  Params[3].Key := 'text';
  Params[3].Value := Text;
  Params[4].Key := 'min_upload_date';
  Params[4].Value := MinUploadDate;
  Params[5].Key := 'max_upload_date';
  Params[5].Value := MaxUploadDate;
  Params[6].Key := 'min_taken_date';
  Params[6].Value := MinTakenDate;
  Params[7].Key := 'max_taken_date';
  Params[7].Value := MaxTakenDate;
  Params[8].Key := 'license';
  Params[8].Value := License;
  Params[9].Key := 'per_page';
  if PerPage > 0 then
    Params[9].Value := IntToStr(PerPage)
  else
    Params[9].Value := '';
  Params[10].Key := 'page';
  if Page > 0 then
    Params[10].Value := IntToStr(Page)
  else
    Params[10].Value := '';
  Params[11].Key := 'sort';
  if (Sort >= 0) and (Sort <= 3) then
    Params[11].Value := FlickrSortLookup[Sort]
  else
    Params[11].Value := '';
  Xml := DoRequest('flickr.photos.search', Params);
  SetLength(Params, 0);

  PhotoList := TFlickrPhotoList.Create(Self);
  try
    PhotoList.GenerateFromXML(Xml, LastError);
  except
    PhotoList.Free;
    raise
  end;
  Result := PhotoList;
end;

// Tags should be a comma separated list with no spaces e.g. 'Tag1,Tag2'
function TFlickr.PhotosAddTags(PhotoId, Tags: string): Boolean;
var
  Params: TFlickrParamArray;
  Xml: string;
begin
  SetLength(Params, 2);
  Params[0].Key := 'photo_id';
  Params[0].Value := PhotoId;
  Params[1].Key := 'tags';
  Params[1].Value := Tags;
  Xml := DoRequest('flickr.photos.addTags', Params);
  SetLength(Params, 0);
  GenerateFromXML(Xml, LastError);
  Result := LastError.Xml = '';
end;

function TFlickr.PeopleFindByEmail(FindEmail: string): TFlickrPeople;
var
  Params: TFlickrParamArray;
  Xml: string;
  People: TFlickrPeople;
begin
  SetLength(Params, 1);
  Params[0].Key := 'find_email';
  Params[0].Value := FindEmail;
  Xml := DoRequest('flickr.people.findByEmail', Params);
  SetLength(Params, 0);

  People := TFlickrPeople.Create(Self);
  try
    People.GenerateFromXML(Xml, LastError);
  except
    People.Free;
    raise
  end;
  Result := People;
end;

function TFlickr.PeopleGetPublicPhotos(UserId: string; PerPage, Page: Integer): TFlickrPhotoList;
var
  Params: TFlickrParamArray;
  Xml: string;
  PhotoList: TFlickrPhotoList;
begin
  SetLength(Params, 3);
  Params[0].Key := 'user_id';
  Params[0].Value := UserId;
  Params[1].Key := 'per_page';
  if PerPage > 0 then
    Params[1].Value := IntToStr(PerPage)
  else
    Params[1].Value := '';
  Params[2].Key := 'page';
  if Page > 0 then
    Params[2].Value := IntToStr(Page)
  else
    Params[2].Value := '';
  Xml := DoRequest('flickr.people.getPublicPhotos', Params);
  SetLength(Params, 0);

  PhotoList := TFlickrPhotoList.Create(Self);
  try
    PhotoList.GenerateFromXML(Xml, LastError);
  except
    PhotoList.Free;
    raise
  end;
  Result := PhotoList;
end;

function TFlickr.PeopleFindByUsername(Username: string): TFlickrPeople;
var
  Params: TFlickrParamArray;
  Xml: string;
  People: TFlickrPeople;
begin
  SetLength(Params, 1);
  Params[0].Key := 'username';
  Params[0].Value := Username;
  Xml := DoRequest('flickr.people.findByUsername', Params);
  SetLength(Params, 0);

  People := TFlickrPeople.Create(Self);
  try
    People.GenerateFromXML(Xml, LastError);
  except
    People.Free;
    raise
  end;
  Result := People;
end;

procedure TFlickr.Parse;
begin
end;

function TFlickr.PeopleGetInfo(People: TFlickrPeople): Boolean;
var
  Params: TFlickrParamArray;
  Xml: string;
begin
  SetLength(Params, 1);
  Params[0].Key := 'user_id';
  Params[0].Value := People.Id;
  Xml := DoRequest('flickr.people.getInfo', Params);
  SetLength(Params, 0);

  People.GenerateFromXML(Xml, LastError);
  Result := LastError.Xml = '';
end;

function TFlickr.PeopleGetOnlineList: TFlickrPeopleList;
var
  Params: TFlickrParamArray;
  Xml: string;
  PeopleList: TFlickrPeopleList;
begin
  SetLength(Params, 0);
  Xml := DoRequest('flickr.people.getOnlineList', Params);

  PeopleList := TFlickrPeopleList.Create(Self);
  try
    PeopleList.GenerateFromXML(Xml, LastError);
  except
    PeopleList.Free;
    raise
  end;
  Result := PeopleList;
end;

function TFlickr.PhotosGetContactsPhotos(Count: Integer; JustFriends, SinglePhoto,
            IncludeSelf: Boolean): TFlickrPhotoList;
var
  Params: TFlickrParamArray;
  Xml: string;
  PhotoList: TFlickrPhotoList;
begin
  SetLength(Params, 4);
  Params[0].Key := 'count';
  if Count > 0 then
    Params[0].Value := IntToStr(Count)
  else
    Params[0].Value := '';
  Params[1].Key := 'just_friends';
  if JustFriends then
    Params[1].Value := '1'
  else
    Params[1].Value := '';
  Params[2].Key := 'single_photo';
  if SinglePhoto then
    Params[2].Value := '1'
  else
    Params[2].Value := '';
  Params[3].Key := 'include_self';
  if IncludeSelf then
    Params[3].Value := '1'
  else
    Params[3].Value := '';
  Xml := DoRequest('flickr.photos.getContactsPhotos', Params);
  SetLength(Params, 0);

  PhotoList := TFlickrPhotoList.Create(Self);
  try
    PhotoList.GenerateFromXML(Xml, LastError);
  except
    PhotoList.Free;
    raise
  end;
  Result := PhotoList;
end;

function TFlickr.PhotosGetContactsPublicPhotos(UserId: string; JustFriends, SinglePhoto,
            IncludeSelf: Boolean; Count: Integer = -1): TFlickrPhotoList;
var
  Params: TFlickrParamArray;
  Xml: string;
  PhotoList: TFlickrPhotoList;
begin
  SetLength(Params, 5);
  Params[0].Key := 'user_id';
  Params[0].Value := UserId;
  Params[1].Key := 'count';
  if Count > 0 then
    Params[1].Value := IntToStr(Count)
  else
    Params[1].Value := '';
  Params[2].Key := 'just_friends';
  if JustFriends then
    Params[2].Value := '1'
  else
    Params[2].Value := '';
  Params[3].Key := 'single_photo';
  if SinglePhoto then
    Params[3].Value := '1'
  else
    Params[3].Value := '';
  Params[4].Key := 'include_self';
  if IncludeSelf then
    Params[4].Value := '1'
  else
    Params[4].Value := '';
  Xml := DoRequest('flickr.photos.getContactsPublicPhotos', Params);
  SetLength(Params, 0);

  PhotoList := TFlickrPhotoList.Create(Self);
  try
    PhotoList.GenerateFromXML(Xml, LastError);
  except
    PhotoList.Free;
    raise
  end;
  Result := PhotoList;
end;

function TFlickr.PhotosGetContext(PhotoId: string; var PrevPhoto,
            NextPhoto: TFlickrPhoto): Boolean;
begin
  LastError.Xml := 'Not implemented yet.';
  Result := False;
end;

function TFlickr.PhotosGetCounts(Dates, TakenDates: string): Boolean;
begin
  LastError.Xml := 'Not implemented yet.';
  Result := False;
end;

function TFlickr.PhotosGetExif(PhotoId, Secret: string): Boolean;
begin
  LastError.Xml := 'Not implemented yet.';
  Result := False;
end;

function TFlickr.PhotosGetInfo(Photo: TFlickrPhoto): Boolean;
var
  Params: TFlickrParamArray;
  Xml: string;
begin
  SetLength(Params, 2);
  Params[0].Key := 'photo_id';
  Params[0].Value := Photo.Id;
  Params[1].Key := 'secret';
  if Photo.Secret <> FLICKR_NULL_STRING then
    Params[1].Value := Photo.Secret
  else
    Params[1].Value := '';
  Xml := DoRequest('flickr.photos.getInfo', Params);
  SetLength(Params, 0);

  Photo.GenerateFromXML(Xml, LastError);
  Result := LastError.Xml = '';
end;

function TFlickr.PhotosGetNotInSet(PerPage, Page: Integer): TFlickrPhotoList;
var
  Params: TFlickrParamArray;
  Xml: string;
  PhotoList: TFlickrPhotoList;
begin
  SetLength(Params, 2);
  Params[0].Key := 'per_page';
  if PerPage > 0 then
    Params[0].Value := IntToStr(PerPage)
  else
    Params[0].Value := '';
  Params[1].Key := 'page';
  if Page > 0 then
    Params[1].Value := IntToStr(Page)
  else
    Params[1].Value := '';
  Xml := DoRequest('flickr.photos.getNotInSet', Params);
  SetLength(Params, 0);

  PhotoList := TFlickrPhotoList.Create(Self);
  try
    PhotoList.GenerateFromXML(Xml, LastError);
  except
    PhotoList.Free;
    raise
  end;
  Result := PhotoList;
end;

function TFlickr.PhotosGetPerms(PhotoId: string): Boolean;
begin
  LastError.Xml := 'Not implemented yet.';
  Result := False;
end;

function TFlickr.PhotosGetRecent(PerPage, Page: Integer): TFlickrPhotoList;
var
  Params: TFlickrParamArray;
  Xml: string;
  PhotoList: TFlickrPhotoList;
begin
  SetLength(Params, 2);
  Params[0].Key := 'per_page';
  if PerPage > 0 then
    Params[0].Value := IntToStr(PerPage)
  else
    Params[0].Value := '';
  Params[1].Key := 'page';
  if Page > 0 then
    Params[1].Value := IntToStr(Page)
  else
    Params[1].Value := '';
  Xml := DoRequest('flickr.photos.getRecent', Params);
  SetLength(Params, 0);

  PhotoList := TFlickrPhotoList.Create(Self);
  try
    PhotoList.GenerateFromXML(Xml, LastError);
  except
    PhotoList.Free;
    raise
  end;
  Result := PhotoList;
end;

function TFlickr.PhotosGetSizes(Photo: TFlickrPhoto): Boolean;
var
  Params: TFlickrParamArray;
  Xml: string;
begin
  SetLength(Params, 1);
  Params[0].Key := 'photo_id';
  Params[0].Value := Photo.Id;
  Xml := DoRequest('flickr.photos.getSizes', Params);
  SetLength(Params, 0);

  // create Photo.FSizes list
  if Photo.FSizes <> nil then Photo.FSizes.Free;
  Photo.FSizes := TFlickrSizeList.Create(Self);
  try
    Photo.FSizes.GenerateFromXML(Xml, LastError);
  except
    Photo.FSizes.Free;
    Photo.FSizes := nil;
    raise
  end;
  Result := LastError.Xml = '';
end;

function TFlickr.PhotosGetUntagged(PerPage, Page: Integer): TFlickrPhotoList;
var
  Params: TFlickrParamArray;
  Xml: string;
  PhotoList: TFlickrPhotoList;
begin
  SetLength(Params, 2);
  Params[0].Key := 'per_page';
  if PerPage > 0 then
    Params[0].Value := IntToStr(PerPage)
  else
    Params[0].Value := '';
  Params[1].Key := 'page';
  if Page > 0 then
    Params[1].Value := IntToStr(Page)
  else
    Params[1].Value := '';
  Xml := DoRequest('flickr.photos.getUntagged', Params);
  SetLength(Params, 0);

  PhotoList := TFlickrPhotoList.Create(Self);
  try
    PhotoList.GenerateFromXML(Xml, LastError);
  except
    PhotoList.Free;
    raise
  end;
  Result := PhotoList;
end;

function TFlickr.PhotosRemoveTag(TagId: string): Boolean;
var
  Params: TFlickrParamArray;
  Xml: string;
begin
  SetLength(Params, 1);
  Params[0].Key := 'tag_id';
  Params[0].Value := TagId;
  Xml := DoRequest('flickr.photos.removeTag', Params);
  SetLength(Params, 0);

  GenerateFromXML(Xml, LastError);
  Result := LastError.Xml = '';
end;

function TFlickr.PhotosSetDates(PhotoId: string; DatePosted: string = '';
            DateTaken: string = ''; DateTakenGranularity: string = ''): Boolean;
var
  Params: TFlickrParamArray;
  Xml: string;
begin
  SetLength(Params, 4);
  Params[0].Key := 'photo_id';
  Params[0].Value := PhotoId;
  Params[1].Key := 'date_posted';
  Params[1].Value := DatePosted;
  Params[2].Key := 'date_taken';
  Params[2].Value := DateTaken;
  Params[3].Key := 'date_taken_granularity';
  Params[3].Value := DateTakenGranularity;
  Xml := DoRequest('flickr.photos.setDates', Params);
  SetLength(Params, 0);

  GenerateFromXML(Xml, LastError);
  Result := LastError.Xml = '';
end;

function TFlickr.PhotosSetMeta(PhotoId, Title, Description: string): Boolean;
var
  Params: TFlickrParamArray;
  Xml: string;
begin
  SetLength(Params, 3);
  Params[0].Key := 'photo_id';
  Params[0].Value := PhotoId;
  Params[1].Key := 'title';
  Params[1].Value := Title;
  Params[2].Key := 'description';
  Params[2].Value := Description;
  Xml := DoRequest('flickr.photos.setMeta', Params);
  SetLength(Params, 0);

  GenerateFromXML(Xml, LastError);
  Result := LastError.Xml = '';
end;

function TFlickr.PhotosSetPerms(PhotoId: string; IsPublic, IsFriend,
  IsFamily: Boolean; PermComment, PermAddMeta: Integer): Boolean;
var
  Params: TFlickrParamArray;
  Xml: string;
begin
  SetLength(Params, 6);
  Params[0].Key := 'photo_id';
  Params[0].Value := PhotoId;
  Params[1].Key := 'is_public';
  if IsPublic then
    Params[1].Value := '1'
  else
    Params[1].Value := '0';
  Params[2].Key := 'is_friend';
  if IsFriend then
    Params[2].Value := '1'
  else
    Params[2].Value := '0';
  Params[3].Key := 'is_family';
  if IsFamily then
    Params[3].Value := '1'
  else
    Params[3].Value := '0';
  Params[4].Key := 'perm_comment';
  Params[4].Value := IntToStr(PermComment);
  Params[5].Key := 'perm_addmeta';
  Params[5].Value := IntToStr(PermAddMeta);
  Xml := DoRequest('flickr.photos.setPerms', Params);
  SetLength(Params, 0);

  GenerateFromXML(Xml, LastError);
  Result := LastError.Xml = '';
end;

// Tags should be a comma separated list with no spaces e.g. 'Tag1,Tag2'
function TFlickr.PhotosSetTags(PhotoId, Tags: string): Boolean;
var
  Params: TFlickrParamArray;
  Xml: string;
begin
  SetLength(Params, 2);
  Params[0].Key := 'photo_id';
  Params[0].Value := PhotoId;
  Params[1].Key := 'tags';
  Params[1].Value := Tags;
  Xml := DoRequest('flickr.photos.setTags', Params);
  SetLength(Params, 0);

  GenerateFromXML(Xml, LastError);
  Result := LastError.Xml = '';
end;

function TFlickr.ContactsGetList: TFlickrPeopleList;
var
  Params: TFlickrParamArray;
  Xml: string;
  PeopleList: TFlickrPeopleList;
begin
  SetLength(Params, 0);
  Xml := DoRequest('flickr.contacts.getList', Params);

  PeopleList := TFlickrPeopleList.Create(Self);
  try
    PeopleList.GenerateFromXML(Xml, LastError);
  except
    PeopleList.Free;
    raise
  end;
  Result := PeopleList;
end;

function TFlickr.ContactsGetPublicList(UserId: string): TFlickrPeopleList;
var
  Params: TFlickrParamArray;
  Xml: string;
  PeopleList: TFlickrPeopleList;
begin
  SetLength(Params, 1);
  Params[0].Key := 'user_id';
  Params[0].Value := UserId;
  Xml := DoRequest('flickr.contacts.getPublicList', Params);
  SetLength(Params, 0);

  PeopleList := TFlickrPeopleList.Create(Self);
  try
    PeopleList.GenerateFromXML(Xml, LastError);
  except
    PeopleList.Free;
    raise
  end;
  Result := PeopleList;
end;

function TFlickr.NotesAdd(PhotoId: string; NoteX, NoteY, NoteW, NoteH: Integer;
            NoteText: string): TFlickrNote;
var
  Params: TFlickrParamArray;
  Xml: string;
  Note: TFlickrNote;
  FlickrXml: TFlickrXml;
begin
  SetLength(Params, 6);
  Params[0].Key := 'photo_id';
  Params[0].Value := PhotoId;
  Params[1].Key := 'note_x';
  Params[1].Value := IntToStr(NoteX);
  Params[2].Key := 'note_y';
  Params[2].Value := IntToStr(NoteY);
  Params[3].Key := 'note_w';
  Params[3].Value := IntToStr(NoteW);
  Params[4].Key := 'note_h';
  Params[4].Value := IntToStr(NoteH);
  Params[5].Key := 'note_text';
  Params[5].Value := NoteText;
  Xml := DoRequest('flickr.photos.notes.add', Params);
  SetLength(Params, 0);

  GenerateFromXML(Xml, LastError);
  Note := nil;
  if LastError.Xml = '' then
  begin
    Note := TFlickrNote.Create(Self);
    FlickrXml := TFlickrXml.Create;
    try
      FlickrXml.GenerateFromXML(Xml, LastError);
      Note.FId := FlickrXml.ParseNotesAdd(Xml);
    except
      FlickrXml.Free;
      Note.Free;
      raise
    end;
    Note.FX := NoteX;
    Note.FY := NoteY;
    Note.FW := NoteW;
    Note.FH := NoteH;
    Note.FText := NoteText;
  end;
  Result := Note;
end;

function TFlickr.NotesEdit(NoteId: string; NoteX, NoteY, NoteW, NoteH: Integer;
            NoteText: string): Boolean;
var
  Params: TFlickrParamArray;
  Xml: string;
begin
  SetLength(Params, 6);
  Params[0].Key := 'note_id';
  Params[0].Value := NoteId;
  Params[1].Key := 'note_x';
  Params[1].Value := IntToStr(NoteX);
  Params[2].Key := 'note_y';
  Params[2].Value := IntToStr(NoteY);
  Params[3].Key := 'note_w';
  Params[3].Value := IntToStr(NoteW);
  Params[4].Key := 'note_h';
  Params[4].Value := IntToStr(NoteH);
  Params[5].Key := 'note_text';
  Params[5].Value := NoteText;
  Xml := DoRequest('flickr.photos.notes.edit', Params);
  SetLength(Params, 0);

  GenerateFromXML(Xml, LastError);
  Result := LastError.Xml = '';
end;

function TFlickr.NotesDelete(NoteId: string): Boolean;
var
  Params: TFlickrParamArray;
  Xml: string;
begin
  SetLength(Params, 1);
  Params[0].Key := 'note_id';
  Params[0].Value := NoteId;
  Xml := DoRequest('flickr.photos.notes.delete', Params);
  SetLength(Params, 0);

  GenerateFromXML(Xml, LastError);
  Result := LastError.Xml = '';
end;

function TFlickr.TagsGetListPhoto(PhotoId: string): Boolean;
begin
  LastError.Xml := 'Not implemented yet.';
  Result := False;
end;

function TFlickr.TagsGetListUserPopular(UserId: string; Count: Integer): TFlickrTagList;
var
  Params: TFlickrParamArray;
  Xml: string;
  TagList: TFlickrTagList;
begin
  SetLength(Params, 2);
  Params[0].Key := 'user_id';
  Params[0].Value := UserId;
  Params[1].Key := 'count';
  Params[1].Value := IntToStr(Count);
  Xml := DoRequest('flickr.tags.getListUserPopular', Params);
  SetLength(Params, 0);

  TagList := TFlickrTagList.Create(Self);
  try
    TagList.GenerateFromXML(Xml, LastError);
  except
    TagList.Free;
    raise
  end;
  Result := TagList;
end;

function TFlickr.TagsGetListUser(UserId: string): TFlickrTagList;
var
  Params: TFlickrParamArray;
  Xml: string;
  TagList: TFlickrTagList;
begin
  SetLength(Params, 1);
  Params[0].Key := 'user_id';
  Params[0].Value := UserId;
  Xml := DoRequest('flickr.tags.getListUser', Params);
  SetLength(Params, 0);

  TagList := TFlickrTagList.Create(Self);
  try
    TagList.GenerateFromXML(Xml, LastError);
  except
    TagList.Free;
    raise
  end;
  Result := TagList;
end;

function TFlickr.FavoritesGetList(UserId: string; PerPage,
            Page: Integer): TFlickrPhotoList;
var
  Params: TFlickrParamArray;
  Xml: string;
  PhotoList: TFlickrPhotoList;
begin
  SetLength(Params, 3);
  Params[0].Key := 'user_id';
  Params[0].Value := UserId;
  Params[1].Key := 'per_page';
  if PerPage > 0 then
    Params[1].Value := IntToStr(PerPage)
  else
    Params[1].Value := '';
  Params[2].Key := 'page';
  if Page > 0 then
    Params[2].Value := IntToStr(Page)
  else
    Params[2].Value := '';
  Xml := DoRequest('flickr.favorites.getList', Params);
  SetLength(Params, 0);

  PhotoList := TFlickrPhotoList.Create(Self);
  try
    PhotoList.GenerateFromXML(Xml, LastError);
  except
    PhotoList.Free;
    raise
  end;
  Result := PhotoList;
end;

function TFlickr.FavoritesGetPublicList(UserId: string; PerPage,
            Page: Integer): TFlickrPhotoList;
var
  Params: TFlickrParamArray;
  Xml: string;
  PhotoList: TFlickrPhotoList;
begin
  SetLength(Params, 3);
  Params[0].Key := 'user_id';
  Params[0].Value := UserId;
  Params[1].Key := 'per_page';
  if PerPage > 0 then
    Params[1].Value := IntToStr(PerPage)
  else
    Params[1].Value := '';
  Params[2].Key := 'page';
  if Page > 0 then
    Params[2].Value := IntToStr(Page)
  else
    Params[2].Value := '';
  Xml := DoRequest('flickr.favorites.getPublicList', Params);
  SetLength(Params, 0);

  PhotoList := TFlickrPhotoList.Create(Self);
  try
    PhotoList.GenerateFromXML(Xml, LastError);
  except
    PhotoList.Free;
    raise
  end;
  Result := PhotoList;
end;

function TFlickr.FavoritesRemove(PhotoId: string): Boolean;
var
  Params: TFlickrParamArray;
  Xml: string;
begin
  SetLength(Params, 1);
  Params[0].Key := 'photo_id';
  Params[0].Value := PhotoId;
  Xml := DoRequest('flickr.favorites.remove', Params);
  SetLength(Params, 0);

  GenerateFromXML(Xml, LastError);
  Result := LastError.Xml = '';
end;

function TFlickr.FavoritesAdd(PhotoId: string): Boolean;
var
  Params: TFlickrParamArray;
  Xml: string;
begin
  SetLength(Params, 1);
  Params[0].Key := 'photo_id';
  Params[0].Value := PhotoId;
  Xml := DoRequest('flickr.favorites.add', Params);
  SetLength(Params, 0);

  GenerateFromXML(Xml, LastError);
  Result := LastError.Xml = '';
end;

function TFlickr.PhotosetsOrderSets(PhotosetIds: string): Boolean;
var
  Params: TFlickrParamArray;
  Xml: string;
begin
  SetLength(Params, 1);
  Params[0].Key := 'photoset_ids';
  Params[0].Value := PhotosetIds;
  Xml := DoRequest('flickr.photosets.orderSets', Params);
  SetLength(Params, 0);
  
  GenerateFromXML(Xml, LastError);
  Result := LastError.Xml = '';
end;

function TFlickr.PhotosetsEditPhotos(PhotosetId, PrimaryPhotoId,
            PhotoIds: string): Boolean;
var
  Params: TFlickrParamArray;
  Xml: string;
begin
  SetLength(Params, 3);
  Params[0].Key := 'photoset_id';
  Params[0].Value := PhotosetId;
  Params[1].Key := 'primary_photo_id';
  Params[1].Value := PrimaryPhotoId;
  Params[2].Key := 'photo_ids';
  Params[2].Value := PhotoIds;
  Xml := DoRequest('flickr.photosets.editPhotos', Params);
  SetLength(Params, 0);
  
  GenerateFromXML(Xml, LastError);
  Result := LastError.Xml = '';
end;

function TFlickr.PhotosetsDelete(PhotosetId: string): Boolean;
var
  Params: TFlickrParamArray;
  Xml: string;
begin
  SetLength(Params, 1);
  Params[0].Key := 'photoset_id';
  Params[0].Value := PhotosetId;
  Xml := DoRequest('flickr.photosets.delete', Params);
  SetLength(Params, 0);

  GenerateFromXML(Xml, LastError);
  Result := LastError.Xml = '';
end;

function TFlickr.PhotosetsGetContext(PhotoId, PhotosetId: string): Boolean;
begin
  LastError.Xml := 'Not implemented.';
  Result := False;
end;

function TFlickr.PhotosetsCreate(Title, Description,
            PrimaryPhotoId: string): TFlickrPhotoSet;
var
  Params: TFlickrParamArray;
  Xml: string;
  Photoset: TFlickrPhotoset;
begin
  SetLength(Params, 3);
  Params[0].Key := 'title';
  Params[0].Value := Title;
  Params[1].Key := 'description';
  Params[1].Value := Description;
  Params[2].Key := 'primary_photo_id';
  Params[2].Value := PrimaryPhotoId;
  Xml := DoRequest('flickr.photosets.create', Params);
  SetLength(Params, 0);

  Photoset := TFlickrPhotoset.Create(Self);
  Photoset.FTitle := Title;
  Photoset.FDescription := Description;
  Photoset.FPrimaryPhotoId := PrimaryPhotoId;
  try
    Photoset.GenerateFromXML(Xml, LastError);
  except
    Photoset.Free;
    raise
  end;
  Result := Photoset;
end;

function TFlickr.PhotosetsGetInfo(Photoset: TFlickrPhotoset): Boolean;
var
  Params: TFlickrParamArray;
  Xml: string;
begin
  SetLength(Params, 1);
  Params[0].Key := 'photoset_id';
  Params[0].Value := Photoset.Id;
  Xml := DoRequest('flickr.photosets.getInfo', Params);
  SetLength(Params, 0);

  Photoset.GenerateFromXML(Xml, LastError);
  Result := LastError.Xml = '';
end;

function TFlickr.PhotosetsGetPhotos(PhotosetId: string): TFlickrPhotoList;
var
  Params: TFlickrParamArray;
  Xml: string;
  PhotoList: TFlickrPhotoList;
begin
  SetLength(Params, 1);
  Params[0].Key := 'photoset_id';
  Params[0].Value := PhotosetId;
  Xml := DoRequest('flickr.photosets.getPhotos', Params);
  SetLength(Params, 0);

  PhotoList := TFlickrPhotoList.Create(Self);
  try
    PhotoList.GenerateFromXML(Xml, LastError);
  except
    PhotoList.Free;
    raise
  end;
  Result := PhotoList;
end;

function TFlickr.PhotosetsGetList(UserId: string): TFlickrPhotosetList;
var
  Params: TFlickrParamArray;
  Xml: string;
  PhotosetList: TFlickrPhotosetList;
begin
  SetLength(Params, 1);
  Params[0].Key := 'user_id';
  Params[0].Value := UserId;
  Xml := DoRequest('flickr.photosets.getList', Params);
  SetLength(Params, 0);

  PhotosetList := TFlickrPhotosetList.Create(Self);
  try
    PhotosetList.GenerateFromXML(Xml, LastError);
  except
    PhotosetList.Free;
    raise
  end;
  Result := PhotosetList;
end;

function TFlickr.PhotosetsEditMeta(PhotosetId, Title,
            Description: string): Boolean;
var
  Params: TFlickrParamArray;
  Xml: string;
begin
  SetLength(Params, 3);
  Params[0].Key := 'photoset_id';
  Params[0].Value := PhotosetId;
  Params[1].Key := 'title';
  Params[1].Value := Title;
  Params[2].Key := 'description';
  Params[2].Value := Description;
  Xml := DoRequest('flickr.photosets.editMeta', Params);
  SetLength(Params, 0);
  
  GenerateFromXML(Xml, LastError);
  Result := LastError.Xml = '';
end;

function TFlickr.BlogsGetList: TFlickrBlogList;
var
  Params: TFlickrParamArray;
  Xml: string;
  BlogList: TFlickrBlogList;
begin
  SetLength(Params, 0);
  Xml := DoRequest('flickr.blogs.getList', Params);

  BlogList := TFlickrBlogList.Create(Self);
  try
    BlogList.GenerateFromXML(Xml, LastError);
  except
    BlogList.Free;
    raise
  end;
  Result := BlogList;
end;

function TFlickr.BlogsPostPhoto(BlogId, PhotoId, Title, Description,
            BlogPassword: string): Boolean;
var
  Params: TFlickrParamArray;
  Xml: string;
begin
  SetLength(Params, 5);
  Params[0].Key := 'blog_id';
  Params[0].Value := BlogId;
  Params[1].Key := 'photo_id';
  Params[1].Value := PhotoId;
  Params[2].Key := 'title';
  Params[2].Value := Title;
  Params[3].Key := 'description';
  Params[3].Value := Description;
  Params[4].Key := 'blog_password';
  Params[4].Value := BlogPassword;
  Xml := DoRequest('flickr.blogs.postPhoto', Params);
  SetLength(Params, 0);

  GenerateFromXML(Xml, LastError);
  Result := LastError.Xml = '';
end;

function TFlickr.GroupsGetInfo(Group: TFlickrGroup): Boolean;
var
  Params: TFlickrParamArray;
  Xml: string;
begin
  SetLength(Params, 1);
  Params[0].Key := 'group_id';
  Params[0].Value := Group.Id;
  Xml := DoRequest('flickr.groups.getInfo', Params);
  SetLength(Params, 0);

  Group.GenerateFromXML(Xml, LastError);
  Result := LastError.Xml = '';
end;

function TFlickr.GroupsGetActiveList: TFlickrGroupList;
var
  Params: TFlickrParamArray;
  Xml: string;
  GroupList: TFlickrGroupList;
begin
  SetLength(Params, 0);
  Xml := DoRequest('flickr.groups.getActiveList', Params);

  GroupList := TFlickrGroupList.Create(Self);
  try
    GroupList.GenerateFromXML(Xml, LastError);
  except
    GroupList.Free;
    raise
  end;
  Result := GroupList;
end;

function TFlickr.GroupsBrowse(CatId: string): TFlickrGroupList;
var
  Params: TFlickrParamArray;
  Xml: string;
  GroupList: TFlickrGroupList;
begin
  SetLength(Params, 1);
  Params[0].Key := 'cat_id';
  Params[0].Value := CatId;
  Xml := DoRequest('flickr.groups.browse', Params);
  SetLength(Params, 0);

  GroupList := TFlickrGroupList.Create(Self);
  try
    GroupList.GenerateFromXML(Xml, LastError);
  except
    GroupList.Free;
    raise
  end;
  Result := GroupList;
end;

function TFlickr.GroupsPoolsGetGroups: TFlickrGroupList;
var
  Params: TFlickrParamArray;
  Xml: string;
  GroupList: TFlickrGroupList;
begin
  SetLength(Params, 0);
  Xml := DoRequest('flickr.groups.pools.getGroups', Params);

  GroupList := TFlickrGroupList.Create(Self);
  try
    GroupList.GenerateFromXML(Xml, LastError);
  except
    GroupList.Free;
    raise
  end;
  Result := GroupList;
end;

function TFlickr.GroupsPoolsGetPhotos(GroupId, Tags: string; PerPage,
            Page: Integer): TFlickrPhotoList;
var
  Params: TFlickrParamArray;
  Xml: string;
  PhotoList: TFlickrPhotoList;
begin
  SetLength(Params, 4);
  Params[0].Key := 'group_id';
  Params[0].Value := GroupId;
  Params[1].Key := 'tags';
  Params[1].Value := Tags;
  Params[2].Key := 'per_page';
  if PerPage > 0 then
    Params[2].Value := IntToStr(PerPage)
  else
    Params[2].Value := '';
  Params[3].Key := 'page';
  if Page > 0 then
    Params[3].Value := IntToStr(Page)
  else
    Params[3].Value := '';
  Xml := DoRequest('flickr.groups.pools.getPhotos', Params);
  SetLength(Params, 0);

  PhotoList := TFlickrPhotoList.Create(Self);
  try
    PhotoList.GenerateFromXML(Xml, LastError);
  except
    PhotoList.Free;
    raise
  end;
  Result := PhotoList;
end;

function TFlickr.GroupsPoolsRemove(PhotoId, GroupId: string): Boolean;
var
  Params: TFlickrParamArray;
  Xml: string;
begin
  SetLength(Params, 2);
  Params[0].Key := 'photo_id';
  Params[0].Value := PhotoId;
  Params[1].Key := 'group_id';
  Params[1].Value := GroupId;
  Xml := DoRequest('flickr.groups.pools.remove', Params);
  SetLength(Params, 0);

  GenerateFromXML(Xml, LastError);
  Result := LastError.Xml = '';
end;

function TFlickr.GroupsPoolsGetContext(PhotoId, GroupId: string): Boolean;
begin
  LastError.Xml := 'Not implemented yet.';
  Result := False;
end;

function TFlickr.GroupsPoolsAdd(PhotoId, GroupId: string): Boolean;
var
  Params: TFlickrParamArray;
  Xml: string;
begin
  SetLength(Params, 2);
  Params[0].Key := 'photo_id';
  Params[0].Value := PhotoId;
  Params[1].Key := 'group_id';
  Params[1].Value := GroupId;
  Xml := DoRequest('flickr.groups.pools.add', Params);
  SetLength(Params, 0);

  GenerateFromXML(Xml, LastError);
  Result := LastError.Xml = '';
end;

function TFlickr.UrlsGetGroup(GroupId: string): string;
var
  Params: TFlickrParamArray;
  Xml: string;
  FlickrXml: TFlickrXml;
begin
  SetLength(Params, 1);
  Params[0].Key := 'group_id';
  Params[0].Value := GroupId;
  Xml := DoRequest('flickr.urls.getGroup', Params);
  SetLength(Params, 0);

  FlickrXml := TFlickrXml.Create;
  try
    FlickrXml.GenerateFromXML(Xml, LastError);
    Result := FlickrXml.ParseUrlsGetGroup(Xml);
  except
    FlickrXml.Free;
    raise
  end;
end;

function TFlickr.UrlsLookupUser(Url: string): string;
var
  Params: TFlickrParamArray;
  Xml: string;
  FlickrXml: TFlickrXml;
begin
  SetLength(Params, 1);
  Params[0].Key := 'url';
  Params[0].Value := Url;
  Xml := DoRequest('flickr.urls.lookupUser', Params);
  SetLength(Params, 0);

  FlickrXml := TFlickrXml.Create;
  try
    FlickrXml.GenerateFromXML(Xml, LastError);
    Result := FlickrXml.ParseUrlsLookupUser(Xml);
  except
    FlickrXml.Free;
    raise
  end;
end;

function TFlickr.UrlsLookupGroup(Url: string): string;
var
  Params: TFlickrParamArray;
  Xml: string;
  FlickrXml: TFlickrXml;
begin
  SetLength(Params, 1);
  Params[0].Key := 'url';
  Params[0].Value := Url;
  Xml := DoRequest('flickr.urls.lookupGroup', Params);
  SetLength(Params, 0);

  FlickrXml := TFlickrXml.Create;
  try
    FlickrXml.GenerateFromXML(Xml, LastError);
    Result := FlickrXml.ParseUrlsLookupGroup(Xml);
  except
    FlickrXml.Free;
    raise
  end;
end;

function TFlickr.UrlsGetUserProfile(UserId: string): string;
var
  Params: TFlickrParamArray;
  Xml: string;
  FlickrXml: TFlickrXml;
begin
  SetLength(Params, 1);
  Params[0].Key := 'user_id';
  Params[0].Value := UserId;
  Xml := DoRequest('flickr.urls.getUserProfile', Params);
  SetLength(Params, 0);

  FlickrXml := TFlickrXml.Create;
  try
    FlickrXml.GenerateFromXML(Xml, LastError);
    Result := FlickrXml.ParseUrlsGetUserUrl(Xml);
  except
    FlickrXml.Free;
    raise
  end;
end;

function TFlickr.UrlsGetUserPhotos(UserId: string): string;
var
  Params: TFlickrParamArray;
  Xml: string;
  FlickrXml: TFlickrXml;
begin
  SetLength(Params, 1);
  Params[0].Key := 'user_id';
  Params[0].Value := UserId;
  Xml := DoRequest('flickr.urls.getUserPhotos', Params);
  SetLength(Params, 0);

  FlickrXml := TFlickrXml.Create;
  try
    FlickrXml.GenerateFromXML(Xml, LastError);
    Result := FlickrXml.ParseUrlsGetUserUrl(Xml);
  except
    FlickrXml.Free;
    raise
  end;
end;

{ TFlickrXml }

procedure TFlickrXml.ParseInitialize(Xml: string);
begin
  FXml := Xml;
  FXmlParser := TXmlParser.Create;
  FXmlParser.LoadFromBuffer(PChar(FXml));
  FXmlParser.Normalize := True;
  FXmlParser.StartScan;
end;

procedure TFlickrXml.ParseFinalize;
begin
  FXmlParser.Free;
end;

// raises exception on error. Also Error is filled.
procedure TFlickrXml.GenerateFromXML(Xml: string; var Error: TFlickrError);
begin
  Error.Xml := '';
  Error.ErrorNo := '';
  Error.ErrorDesc := '';
  // check for failed call
  if Pos('<rsp stat="fail">', Xml) > 0 then
  begin
    Error := ParseError(StringReplace(Xml, #10, #13#10, [rfReplaceAll]));
    raise EFlickrError.Create(Error.ErrorNo + '=' + Error.ErrorDesc);
  end
  else
  begin
    ParseInitialize(Xml);
    Parse;
    ParseFinalize;
  end;
end;

// Parse flickr.photos.notes.add response
function TFlickrXml.ParseNotesAdd(Xml: string): string;
begin
  Result := '';
  ParseInitialize(Xml);
  while FXmlParser.Scan do
  begin
    case FXmlParser.CurPartType of
      ptStartTag,
      ptEmptyTag:
        begin
          // read attributes
          if (FXmlParser.CurName = 'note') then
            if (FXmlParser.CurAttr.Count > 0) then
              SetAttr('id', Result);
        end;
      else
        begin
        end;
    end;
  end;
  ParseFinalize;
end;

// check if attribute exists and if yes set it
procedure TFlickrXml.SetAttr(AttrName: string; var Variable: Integer);
var
  Node: TNvpNode;
begin
  Node := FXmlParser.CurAttr.Node(AttrName);
  if Node <> nil then
    try Variable := StrToInt(Node.Value); except Variable := -1; end;
end;

// check if attribute exists and if yes set it
procedure TFlickrXml.SetAttr(AttrName: string; var Variable: string);
var
  Node: TNvpNode;
begin
  Node := FXmlParser.CurAttr.Node(AttrName);
  if Node <> nil then
    Variable := Node.Value;
end;

procedure TFlickrXml.Parse;
begin
end;

function TFlickrXml.ParseUrlsGetGroup(Xml: string): string;
begin
  Result := '';
  ParseInitialize(Xml);
  while FXmlParser.Scan do
  begin
    case FXmlParser.CurPartType of
      ptStartTag,
      ptEmptyTag:
        begin
          // read attributes
          if (FXmlParser.CurName = 'group') then
            if (FXmlParser.CurAttr.Count > 0) then
              SetAttr('url', Result);
        end;
      else
        begin
        end;
    end;
  end;
  ParseFinalize;
end;

function TFlickrXml.ParseUrlsGetUserUrl(Xml: string): string;
begin
  Result := '';
  ParseInitialize(Xml);
  while FXmlParser.Scan do
  begin
    case FXmlParser.CurPartType of
      ptStartTag,
      ptEmptyTag:
        begin
          // read attributes
          if (FXmlParser.CurName = 'user') then
            if (FXmlParser.CurAttr.Count > 0) then
              SetAttr('url', Result);
        end;
      else
        begin
        end;
    end;
  end;
  ParseFinalize;
end;

function TFlickrXml.ParseUrlsLookupGroup(Xml: string): string;
begin
  Result := '';
  ParseInitialize(Xml);
  while FXmlParser.Scan do
  begin
    case FXmlParser.CurPartType of
      ptContent,
      ptCData:
        begin
          if (FXmlParser.CurName = 'groupname') then
            Result := FXmlParser.CurContent;
        end;
      else
        begin
        end;
    end;
  end;
  ParseFinalize;
end;

function TFlickrXml.ParseUrlsLookupUser(Xml: string): string;
begin
  Result := '';
  ParseInitialize(Xml);
  while FXmlParser.Scan do
  begin
    case FXmlParser.CurPartType of
      ptContent,
      ptCData:
        begin
          if (FXmlParser.CurName = 'username') then
            Result := FXmlParser.CurContent;
        end;
      else
        begin
        end;
    end;
  end;
  ParseFinalize;
end;

function TFlickrXml.ParseError(Xml: string): TFlickrError;
begin
  Result.Xml := Xml;
  ParseInitialize(Xml);
  while FXmlParser.Scan do
  begin
    case FXmlParser.CurPartType of
      ptStartTag,
      ptEmptyTag:
        begin
          // read attributes
          if (FXmlParser.CurName = 'err') then
            if (FXmlParser.CurAttr.Count > 0) then
            begin
              SetAttr('code', Result.ErrorNo);
              SetAttr('msg', Result.ErrorDesc);
            end;
        end;
      else
        begin
        end;
    end;
  end;
  ParseFinalize;
end;

{ TFlickrPeople }

constructor TFlickrPeople.Create(Flickr: TFlickr);
begin
  FFlickr := Flickr;
  FId := FLICKR_NULL_STRING;
  FContactsPublicPhotos := nil;
  FOnline := FLICKR_NULL_STRING;
  FAwayMessage := FLICKR_NULL_STRING;
  ResetProperties;
end;

destructor TFlickrPeople.Destroy;
begin
  if FPublicPhotos <> nil then
    FPublicPhotos.Free;
  if FContactsPublicPhotos <> nil then
    FContactsPublicPhotos.Free;
  if FPublicContacts <> nil then
    FPublicContacts.Free;
  if FTags <> nil then
    FTags.Free;
  if FTagsPopular <> nil then
    FTagsPopular.Free;
  if FFavorites <> nil then
    FFavorites.Free;
  if FFavoritesPublic <> nil then
    FFavoritesPublic.Free;
  if FPhotosets <> nil then
    FPhotosets.Free;
  inherited;
end;

function TFlickrPeople.GetPhotosCount: Integer;
begin
  if FPhotosCount = FLICKR_NULL_NUMBER then
    FFlickr.PeopleGetInfo(Self);
  Result := FPhotosCount;
end;

function TFlickrPeople.GetLocation: string;
begin
  if FLocation = FLICKR_NULL_STRING then
    FFlickr.PeopleGetInfo(Self);
  Result := FLocation;
end;

function TFlickrPeople.GetPhotosFirstDateTaken: string;
begin
  if FPhotosFirstDateTaken = FLICKR_NULL_STRING then
    FFlickr.PeopleGetInfo(Self);
  Result := FPhotosFirstDateTaken;
end;

function TFlickrPeople.GetIsPro: Boolean;
begin
  if FIsPro = FLICKR_NULL_STRING then
    FFlickr.PeopleGetInfo(Self);
  Result := FIsPro = '1';
end;

function TFlickrPeople.GetPhotosFirstDate: string;
begin
  if FPhotosFirstDate = FLICKR_NULL_STRING then
    FFlickr.PeopleGetInfo(Self);
  Result := FPhotosFirstDate;
end;

function TFlickrPeople.GetRealname: string;
begin
  if FRealname = FLICKR_NULL_STRING then
    FFlickr.PeopleGetInfo(Self);
  Result := FRealname;
end;

function TFlickrPeople.GetUsername: string;
begin
  if FUsername = FLICKR_NULL_STRING then
    FFlickr.PeopleGetInfo(Self);
  Result := FUsername;
end;

function TFlickrPeople.GetIsAdmin: Boolean;
begin
  if FIsAdmin = FLICKR_NULL_STRING then
    FFlickr.PeopleGetInfo(Self);
  Result := FIsAdmin = '1';
end;

// return List of public photos
function TFlickrPeople.GetPublicPhotos: TFlickrPhotoList;
begin
  if FPublicPhotos = nil then
    FPublicPhotos := FFlickr.PeopleGetPublicPhotos(Id, -1, -1);
  Result := FPublicPhotos;
end;

// return List of public contacts
function TFlickrPeople.GetPublicContacts: TFlickrPeopleList;
begin
  if FPublicContacts = nil then
    FPublicContacts := FFlickr.ContactsGetPublicList(Id);
  Result := FPublicContacts;
end;

// return list of tags
function TFlickrPeople.GetTags: TFlickrTagList;
begin
  if FTags = nil then
    FTags := FFlickr.TagsGetListUser(FId);
  Result := FTags;
end;

// return list of popular tags
function TFlickrPeople.GetTagsPopular: TFlickrTagList;
begin
  if FTagsPopular = nil then
    FTagsPopular := FFlickr.TagsGetListUserPopular(FId, 10);
  Result := FTagsPopular;
end;

// return list of public favorites
function TFlickrPeople.GetFavoritesPublic: TFlickrPhotoList;
begin
  if FFavoritesPublic = nil then
    FFavoritesPublic := FFlickr.FavoritesGetPublicList(Fid, -1, -1);
  Result := FFavoritesPublic;
end;

// return list of favorites
function TFlickrPeople.GetFavorites: TFlickrPhotoList;
begin
  if FFavorites = nil then
    FFavorites := FFlickr.FavoritesGetList(Fid, -1, -1);
  Result := FFavorites;
end;

// return list of photosets
function TFlickrPeople.GetPhotosets: TFlickrPhotosetList;
begin
  if FPhotosets = nil then
    FPhotosets := FFlickr.PhotosetsGetList(FId);
  Result := FPhotosets;
end;

procedure TFlickrPeople.Parse;
begin
  while FXmlParser.Scan do
  begin
    case FXmlParser.CurPartType of
      ptStartTag,
      ptEmptyTag:
        begin
          // read attributes
          if (FXmlParser.CurName = 'user') then
          begin
            if (FXmlParser.CurAttr.Count > 0) then
            begin
              SetAttr('nsid', FId);
            end;
          end;
          if (FXmlParser.CurName = 'person') then
          begin
            if (FXmlParser.CurAttr.Count > 0) then
            begin
              SetAttr('nsid', FId);
              SetAttr('isadmin', FIsAdmin);
              SetAttr('ispro', FIsPro);
            end;
          end;
          // initialize contents
          if (FXmlParser.CurName = 'username') then
            FUsername := '';
          if (FXmlParser.CurName = 'realname') then
            FRealname := '';
          if (FXmlParser.CurName = 'location') then
            FLocation := '';
          if (FXmlParser.CurName = 'firstdate') then
            FPhotosFirstDate := '';
          if (FXmlParser.CurName = 'firstdatetaken') then
            FPhotosFirstDateTaken := '';
          if (FXmlParser.CurName = 'count') then
            FPhotosCount := -1;
        end;
      ptContent,
      ptCData:
        begin
          if (FXmlParser.CurName = 'username') then
            FUsername := FXmlParser.CurContent;
          if (FXmlParser.CurName = 'realname') then
            FRealname := FXmlParser.CurContent;
          if (FXmlParser.CurName = 'location') then
            FLocation := FXmlParser.CurContent;
          if (FXmlParser.CurName = 'firstdate') then
            FPhotosFirstDate := FXmlParser.CurContent;
          if (FXmlParser.CurName = 'firstdatetaken') then
            FPhotosFirstDateTaken := FXmlParser.CurContent;
          if (FXmlParser.CurName = 'count') then
            try FPhotosCount := StrToInt(FXmlParser.CurContent); except FPhotosCount := -1; end;
        end;
      else
        begin
        end;
    end;
  end;
end;

function TFlickrPeople.GetUrlProfile: string;
begin
  if FUrlProfile = FLICKR_NULL_STRING then
    FUrlProfile := FFlickr.UrlsGetUserProfile(FId);
  Result := FUrlProfile;
end;

function TFlickrPeople.GetUrlPhotos: string;
begin
  if FUrlPhotos = FLICKR_NULL_STRING then
    FUrlPhotos := FFlickr.UrlsGetUserPhotos(FId);
  Result := FUrlPhotos;
end;

function TFlickrPeople.GetContactsPublicPhotos(PJustFriends, PSinglePhoto,
            PIncludeSelf: Boolean; PCount: Integer = -1): TFlickrPhotoList;
begin
  if FContactsPublicPhotos = nil then
    FContactsPublicPhotos := FFlickr.PhotosGetContactsPublicPhotos(FId,
      PJustFriends, PSinglePhoto, PIncludeSelf, PCount);
  Result := FContactsPublicPhotos;
end;

procedure TFlickrPeople.ResetProperties;
begin
  FIsAdmin := FLICKR_NULL_STRING;
  FIsPro := FLICKR_NULL_STRING;
  FUsername := FLICKR_NULL_STRING;
  FRealname := FLICKR_NULL_STRING;
  FLocation := FLICKR_NULL_STRING;
  FPhotosFirstDate := FLICKR_NULL_STRING;
  FPhotosFirstDateTaken := FLICKR_NULL_STRING;
  FPhotosCount := FLICKR_NULL_NUMBER;
  FUrlPhotos := FLICKR_NULL_STRING;
  FUrlProfile := FLICKR_NULL_STRING;
  if FPublicPhotos <> nil then FPublicPhotos.Free;
  FPublicPhotos := nil;
  if FPublicContacts <> nil then FPublicPhotos.Free;
  FPublicContacts := nil;
  if FTags <> nil then FTags.Free;
  FTags := nil;
  if FTagsPopular <> nil then FTagsPopular.Free;
  FTagsPopular := nil;
  if FFavorites <> nil then FFavorites.Free;
  FFavorites := nil;
  if FFavoritesPublic <> nil then FFavoritesPublic.Free;
  FFavoritesPublic := nil;
  if FPhotosets <> nil then FPhotosets.Free;
  FPhotosets := nil;
  if FContactsPublicPhotos <> nil then FContactsPublicPhotos.Free;
  FContactsPublicPhotos := nil;
end;

{ TFlickrPeopleList }

procedure TFlickrPeopleList.Parse;
var
  People: TFlickrPeople;
begin
  People := nil;
  while FXmlParser.Scan do
  begin
    case FXmlParser.CurPartType of
      ptStartTag,
      ptEmptyTag:
        begin
          // read attributes
          if (FXmlParser.CurName = 'user') or (FXmlParser.CurName = 'contact') then
          begin
            if (FXmlParser.CurAttr.Count > 0) then
            begin
              People := TFlickrPeople.Create(FFlickr);
              SetAttr('nsid', People.FId);
              SetAttr('username', People.FUsername);
              SetAttr('realname', People.FRealname);
              SetAttr('online', People.FOnline);
              // TODO: Read Friend/Family/Ignored
              FPeopleList.Add(People);
            end;
          end;
          // initialize contents
          if ((FXmlParser.CurName = 'user') or (FXmlParser.CurName = 'contact')) and (People <> nil) then
            People.FAwayMessage := '';
        end;
      ptContent,
      ptCData:
        begin
          if ((FXmlParser.CurName = 'user') or (FXmlParser.CurName = 'contact')) and (People <> nil) then
            People.FAwayMessage := FXmlParser.CurContent;
        end;
      else
        begin
        end;
    end;
  end;
end;

constructor TFlickrPeopleList.Create(Flickr: TFlickr);
begin
  FFlickr := Flickr;
  FPeopleList := TObjectList.Create;
end;

function TFlickrPeopleList.Count: Integer;
begin
  Result := FPeopleList.Count;
end;

function TFlickrPeopleList.GetPeople(i: Integer): TFlickrPeople;
begin
  Result := TFlickrPeople(FPeopleList[i]);
end;

destructor TFlickrPeopleList.Destroy;
begin
  FPeopleList.Free;
  inherited;
end;

{ TFlickrPhoto }

constructor TFlickrPhoto.Create(Flickr: TFlickr);
begin
  FFlickr := Flickr;
  FId := FLICKR_NULL_STRING;
  FIsPrimary := FLICKR_NULL_STRING;
  FSecret := FLICKR_NULL_STRING; // i don't want secret changed by ResetProperties
  ResetProperties;
end;

function TFlickrPhoto.GetOwnerLocation: string;
begin
  if FOwnerLocation = FLICKR_NULL_STRING then
    FFlickr.PhotosGetInfo(Self);
  Result := FOwnerLocation;
end;

function TFlickrPhoto.GetPermComment: string;
begin
  if FPermComment = FLICKR_NULL_STRING then
    FFlickr.PhotosGetInfo(Self);
  Result := FPermComment;
end;

function TFlickrPhoto.GetIsFamily: Boolean;
begin
  if FIsFamily = FLICKR_NULL_STRING then
    FFlickr.PhotosGetInfo(Self);
  Result := FIsFamily = '1';
end;

function TFlickrPhoto.GetDatePosted: string;
begin
  if FDatePosted = FLICKR_NULL_STRING then
    FFlickr.PhotosGetInfo(Self);
  Result := FDatePosted;
end;

function TFlickrPhoto.GetTitle: string;
begin
  if FTitle = FLICKR_NULL_STRING then
    FFlickr.PhotosGetInfo(Self);
  Result := FTitle;
end;

function TFlickrPhoto.GetSecret: string;
begin
  if FSecret = FLICKR_NULL_STRING then
    FFlickr.PhotosGetInfo(Self);
  Result := FSecret;
end;

function TFlickrPhoto.GetIsFavorite: Boolean;
begin
  if FIsFavorite = FLICKR_NULL_STRING then
    FFlickr.PhotosGetInfo(Self);
  Result := FIsFavorite = '1';
end;

function TFlickrPhoto.GetDescription: string;
begin
  if FDescription = FLICKR_NULL_STRING then
    FFlickr.PhotosGetInfo(Self);
  Result := FDescription;
end;

function TFlickrPhoto.GetCanAddMeta: string;
begin
  if FCanAddMeta = FLICKR_NULL_STRING then
    FFlickr.PhotosGetInfo(Self);
  Result := FCanAddMeta;
end;

function TFlickrPhoto.GetOwnerId: string;
begin
  if FOwnerId = FLICKR_NULL_STRING then
    FFlickr.PhotosGetInfo(Self);
  Result := FOwnerId;
end;

function TFlickrPhoto.GetDateTaken: string;
begin
  if FDateTaken = FLICKR_NULL_STRING then
    FFlickr.PhotosGetInfo(Self);
  Result := FDateTaken;
end;

function TFlickrPhoto.GetCanComment: string;
begin
  if FCanComment = FLICKR_NULL_STRING then
    FFlickr.PhotosGetInfo(Self);
  Result := FCanComment;
end;

function TFlickrPhoto.GetIsFriend: Boolean;
begin
  if FIsFriend = FLICKR_NULL_STRING then
    FFlickr.PhotosGetInfo(Self);
  Result := FIsFriend = '1';
end;

function TFlickrPhoto.GetServer: string;
begin
  if FServer = FLICKR_NULL_STRING then
    FFlickr.PhotosGetInfo(Self);
  Result := FServer;
end;

function TFlickrPhoto.GetOwnerRealname: string;
begin
  if FOwnerRealname = FLICKR_NULL_STRING then
    FFlickr.PhotosGetInfo(Self);
  Result := FOwnerRealname;
end;

function TFlickrPhoto.GetDateTakenGranularity: string;
begin
  if FDateTakenGranularity = FLICKR_NULL_STRING then
    FFlickr.PhotosGetInfo(Self);
  Result := FDateTakenGranularity;
end;

function TFlickrPhoto.GetPermAddMeta: string;
begin
  if FPermAddMeta = FLICKR_NULL_STRING then
    FFlickr.PhotosGetInfo(Self);
  Result := FPermAddMeta;
end;

function TFlickrPhoto.GetOwnerUsername: string;
begin
  if FOwnerUsername = FLICKR_NULL_STRING then
    FFlickr.PhotosGetInfo(Self);
  Result := FOwnerUsername;
end;

function TFlickrPhoto.GetLicense: string;
begin
  if FLicense = FLICKR_NULL_STRING then
    FFlickr.PhotosGetInfo(Self);
  Result := FLicense;
end;

function TFlickrPhoto.GetIsPublic: Boolean;
begin
  if FIsPublic = FLICKR_NULL_STRING then
    FFlickr.PhotosGetInfo(Self);
  Result := FIsPublic = '1';
end;

function TFlickrPhoto.GetCommentsCount: Integer;
begin
  if FCommentsCount = FLICKR_NULL_NUMBER then
    FFlickr.PhotosGetInfo(Self);
  Result := FCommentsCount;
end;

function TFlickrPhoto.GetNotes: TFlickrNoteList;
begin
  if FNotes = nil then
    FFlickr.PhotosGetInfo(Self);
  Result := FNotes;
end;

function TFlickrPhoto.GetTags: TFlickrTagList;
begin
  if FTags = nil then
    FFlickr.PhotosGetInfo(Self);
  Result := FTags;
end;


procedure TFlickrPhoto.Parse;
var
  Note: TFlickrNote;
  Tag: TFlickrTag;
begin
  Note := nil;
  Tag := nil;
  // Notes and Tags must be created here (for cache to work)
  if FNotes <> nil then FNotes.Free;
  FNotes := TFlickrNoteList.Create(FFlickr);
  if FTags <> nil then FTags.Free;
  FTags := TFlickrTagList.Create(FFlickr);

  while FXmlParser.Scan do
  begin
    case FXmlParser.CurPartType of
      ptStartTag,
      ptEmptyTag:
        begin
          // read attributes
          if (FXmlParser.CurName = 'photo') then
          begin
            if (FXmlParser.CurAttr.Count > 0) then
            begin
              SetAttr('nsid', FId);
              SetAttr('secret', FSecret);
              SetAttr('server', FServer);
              SetAttr('isfavorite', FIsFavorite);
              SetAttr('license', FLicense);
            end;
          end;
          if (FXmlParser.CurName = 'owner') then
          begin
            if (FXmlParser.CurAttr.Count > 0) then
            begin
              SetAttr('nsid', FOwnerId);
              SetAttr('username', FOwnerUsername);
              SetAttr('realname', FOwnerRealname);
              SetAttr('location', FOwnerLocation);
            end;
          end;
          if (FXmlParser.CurName = 'visibility') then
          begin
            if (FXmlParser.CurAttr.Count > 0) then
            begin
              SetAttr('is_public', FIsPublic);
              SetAttr('is_friend', FIsFriend);
              SetAttr('is_family', FIsFamily);
            end;
          end;
          if (FXmlParser.CurName = 'dates') then
          begin
            if (FXmlParser.CurAttr.Count > 0) then
            begin
              SetAttr('posted', FDatePosted);
              SetAttr('taken', FDateTaken);
              SetAttr('takengranularity', FDateTakenGranularity);
            end;
          end;
          if (FXmlParser.CurName = 'permissions') then
          begin
            if (FXmlParser.CurAttr.Count > 0) then
            begin
              SetAttr('permcomment', FPermComment);
              SetAttr('permaddmeta', FPermAddMeta);
            end;
          end;
          if (FXmlParser.CurName = 'editability') then
          begin
            if (FXmlParser.CurAttr.Count > 0) then
            begin
              SetAttr('cancomment', FCanComment);
              SetAttr('canaddmeta', FCanAddMeta);
            end;
          end;
          if (FXmlParser.CurName = 'note') then
          begin
            if (FXmlParser.CurAttr.Count > 0) then
            begin
              Note := TFlickrNote.Create(FFlickr);
              FNotes.FNoteList.Add(Note);
              SetAttr('id', Note.FId);
              SetAttr('author', Note.FAuthorId);
              SetAttr('authorname', Note.FAuthorName);
              SetAttr('x', Note.FX);
              SetAttr('y', Note.FY);
              SetAttr('w', Note.FW);
              SetAttr('h', Note.FH);
            end;
          end;
          if (FXmlParser.CurName = 'tag') then
          begin
            if (FXmlParser.CurAttr.Count > 0) then
            begin
              Tag := TFlickrTag.Create(FFlickr);
              FTags.FTagList.Add(Tag);
              SetAttr('id', Tag.FId);
              SetAttr('author', Tag.FAuthorId);
              SetAttr('raw', Tag.FText);
            end;
          end;
          // initialize contents
          if (FXmlParser.CurName = 'title') then
            FTitle := '';
          if (FXmlParser.CurName = 'description') then
            FDescription := '';
          if (FXmlParser.CurName = 'comments') then
            FCommentsCount := -1;
          if (FXmlParser.CurName = 'note') and (Note <> nil) then
            Note.FText := '';
          if (FXmlParser.CurName = 'tag') and (Tag <> nil) and (Tag.FText = FLICKR_NULL_STRING) then
            Tag.FText := '';
        end;
      ptContent,
      ptCData:
        begin
          if (FXmlParser.CurName = 'title') then
            FTitle := FXmlParser.CurContent;
          if (FXmlParser.CurName = 'description') then
            FDescription := FXmlParser.CurContent;
          if (FXmlParser.CurName = 'comments') then
            try FCommentsCount := StrToInt(FXmlParser.CurContent); except FCommentsCount := -1; end;
          if (FXmlParser.CurName = 'note') and (Note <> nil) then
            Note.FText := FXmlParser.CurContent;
          if ((FXmlParser.CurName = 'tag') and (Tag <> nil) and ((Tag.FText = '') or (Tag.FText = FLICKR_NULL_STRING))) then
            Tag.FText := FXmlParser.CurContent;
        end;
      else
        begin
        end;
    end;
  end;
end;

destructor TFlickrPhoto.Destroy;
begin
  if FNotes <> nil then
    FNotes.Free;
  if FTags <> nil then
    FTags.Free;
  if FSizes <> nil then
    FSizes.Free;
  inherited;
end;

function TFlickrPhoto.GetSizes: TFlickrSizeList;
begin
  if FSizes = nil then
    FFlickr.PhotosGetSizes(Self);
  Result := FSizes;
end;

function TFlickrPhoto.NotesAdd(X, Y, W, H: Integer; Text: string): TFlickrNote;
begin
  Result := Notes.AddNote(FId, X, Y, W, H, Text);
end;

function TFlickrPhoto.NotesEdit(NoteId: string; X, Y, W, H: Integer; Text: string): Boolean;
begin
  Result := Notes.EdtNote(NoteId, X, Y, W, H, Text);
end;

function TFlickrPhoto.NotesDelete(NoteId: string): Boolean;
begin
  Result := Notes.DelNote(NoteId);
end;

function TFlickrPhoto.FavoritesRemove: Boolean;
begin
  Result := FFlickr.FavoritesRemove(FId);
end;

function TFlickrPhoto.FavoritesAdd: Boolean;
begin
  Result := FFlickr.FavoritesAdd(FId);
end;

function TFlickrPhoto.TagsAdd(PTags: string): Boolean;
begin
  Result := FFlickr.PhotosAddTags(FId, PTags);
end;

function TFlickrPhoto.SetDates(PDatePosted: string = '';
            PDateTaken: string = ''; PDateTakenGranularity: string = ''): Boolean;
begin
  Result := FFlickr.PhotosSetDates(FId, PDatePosted, PDateTaken, PDateTakenGranularity);
end;

function TFlickrPhoto.SetTags(PTags: string): Boolean;
begin
  Result := FFlickr.PhotosSetTags(FId, PTags);
end;

function TFlickrPhoto.SetMeta(PTitle, PDescription: string): Boolean;
begin
  Result := FFlickr.PhotosSetMeta(FId, PTitle, PDescription);
end;

function TFlickrPhoto.SetPerms(PIsPublic, PIsFriend, PIsFamily: Boolean;
            PPermComment, PPermAddMeta: Integer): Boolean;
begin
  Result := FFlickr.PhotosSetPerms(FId, PIsPublic, PIsFriend, PIsFamily, PPermComment, PPermAddMeta);
end;

// All properties that get cached, should be emptied here
procedure TFlickrPhoto.ResetProperties;
begin
  FOwnerId := FLICKR_NULL_STRING;
  FOwnerUsername := FLICKR_NULL_STRING;
  FOwnerRealname := FLICKR_NULL_STRING;
  FOwnerLocation := FLICKR_NULL_STRING;
  FServer := FLICKR_NULL_STRING;
  FTitle := FLICKR_NULL_STRING;
  FDescription := FLICKR_NULL_STRING;
  FIsPublic := FLICKR_NULL_STRING;
  FIsFriend := FLICKR_NULL_STRING;
  FIsFamily := FLICKR_NULL_STRING;
  FIsFavorite := FLICKR_NULL_STRING;
  FLicense := FLICKR_NULL_STRING;
  FDatePosted := FLICKR_NULL_STRING;
  FDateTaken := FLICKR_NULL_STRING;
  FDateTakenGranularity := FLICKR_NULL_STRING;
  FPermComment := FLICKR_NULL_STRING;
  FPermAddMeta := FLICKR_NULL_STRING;
  FCanComment := FLICKR_NULL_STRING;
  FCanAddMeta := FLICKR_NULL_STRING;
  FCommentsCount := FLICKR_NULL_NUMBER;
  if FNotes <> nil then FNotes.Free;
  FNotes := nil;
  if FTags <> nil then FTags.Free;
  FTags := nil;
  if FSizes <> nil then FSizes.Free;
  FSizes := nil;
end;

{ TFlickrNote }

constructor TFlickrNote.Create(Flickr: TFlickr);
begin
  FFlickr := Flickr;
end;

{ TFlickrNoteList }

constructor TFlickrNoteList.Create(Flickr: TFlickr);
begin
  FFlickr := Flickr;
  FNoteList := TObjectList.Create;
end;

function TFlickrNoteList.GetNote(i: Integer): TFlickrNote;
begin
  Result := TFlickrNote(FNoteList[i]);
end;

function TFlickrNoteList.Count: Integer;
begin
  Result := FNoteList.Count;
end;

destructor TFlickrNoteList.Destroy;
begin
  FNoteList.Free;
  inherited;
end;

function TFlickrNoteList.AddNote(PhotoId: string; X, Y, W, H: Integer; Text: string): TFlickrNote;
var
  Note: TFlickrNote;
begin
  Note := FFlickr.NotesAdd(PhotoId, X, Y, W, H, Text);
  if Note <> nil then
    FNoteList.Add(Note);
  Result := Note;
end;

function TFlickrNoteList.EdtNote(NoteId: string; X, Y, W, H: Integer; Text: string): Boolean;
begin
  Result := FFlickr.NotesEdit(NoteId, X, Y, W, H, Text);
end;

function TFlickrNoteList.DelNote(NoteId: string): Boolean;
begin
  Result := FFlickr.NotesDelete(NoteId);
end;

{ TFlickrTag }

constructor TFlickrTag.Create(Flickr: TFlickr);
begin
  FFlickr := Flickr;
end;

{ TFlickrTagList }

constructor TFlickrTagList.Create(Flickr: TFlickr);
begin
  FFlickr := Flickr;
  FTagList := TObjectList.Create;
end;

function TFlickrTagList.GetTag(i: Integer): TFlickrTag;
begin
  Result := TFlickrTag(FTagList[i]);
end;

function TFlickrTagList.Count: Integer;
begin
  Result := FTagList.Count;
end;

destructor TFlickrTagList.Destroy;
begin
  FTagList.Free;
  inherited;
end;

procedure TFlickrTagList.Parse;
var
  AuthorId: string;
  Tag: TFlickrTag;
begin
  Tag := nil;
  AuthorId := '';
  while FXmlParser.Scan do
  begin
    case FXmlParser.CurPartType of
      ptStartTag,
      ptEmptyTag:
        begin
          // read attributes
          if (FXmlParser.CurName = 'who') then
          begin
            if (FXmlParser.CurAttr.Count > 0) then
              SetAttr('id', AuthorId);
          end;
          if (FXmlParser.CurName = 'tag') then
          begin
            Tag := TFlickrTag.Create(FFlickr);
            Tag.FAuthorId := AuthorId;
            FTagList.Add(Tag);
            if (FXmlParser.CurAttr.Count > 0) then
            begin
              SetAttr('count', Tag.FCount);
              SetAttr('raw', Tag.FText);
            end;
          end;
          // initialize contents
          if ((FXmlParser.CurName = 'tag') and (Tag <> nil) and (Tag.FText = FLICKR_NULL_STRING)) then
            Tag.FText := '';
        end;
      ptContent,
      ptCData:
        begin
          if ((FXmlParser.CurName = 'tag') and (Tag <> nil) and ((Tag.FText = '') or (Tag.FText = FLICKR_NULL_STRING))) then
            Tag.FText := FXmlParser.CurContent;
        end;
      else
        begin
        end;
    end;
  end;
end;

{ TFlickrSize }

constructor TFlickrSize.Create(Flickr: TFlickr);
begin
  FFlickr := Flickr;
end;

{ TFlickrSizeList }

constructor TFlickrSizeList.Create(Flickr: TFlickr);
begin
  FFlickr := Flickr;
  FSizeList := TObjectList.Create;
end;

function TFlickrSizeList.Count: Integer;
begin
  Result := FSizeList.Count;
end;

function TFlickrSizeList.GetSize(i: Integer): TFlickrSize;
begin
  Result := TFlickrSize(FSizeList[i]);
end;

destructor TFlickrSizeList.Destroy;
begin
  FSizeList.Free;
  inherited;
end;

procedure TFlickrSizeList.Parse;
var
  Size: TFlickrSize;
begin
  while FXmlParser.Scan do
  begin
    case FXmlParser.CurPartType of
      ptStartTag,
      ptEmptyTag:
        begin
          // read attributes
          if (FXmlParser.CurName = 'size') then
          begin
            if (FXmlParser.CurAttr.Count > 0) then
            begin
              Size := TFlickrSize.Create(FFlickr);
              FSizeList.Add(Size);
              SetAttr('label', Size.FText);
              SetAttr('source', Size.FSource);
              SetAttr('url', Size.FUrl);
              SetAttr('width', Size.FWidth);
              SetAttr('height', Size.FHeight);
            end;
          end;
        end;
      ptContent,
      ptCData:
        begin
        end;
      else
        begin
        end;
    end;
  end;
end;

{ TFlickrPhotoset }

constructor TFlickrPhotoset.Create(Flickr: TFlickr);
begin
  FFlickr := Flickr;
  FId := FLICKR_NULL_STRING;
  FUrl := FLICKR_NULL_STRING;
  ResetProperties;
end;

destructor TFlickrPhotoset.Destroy;
begin
  if FPhotos <> nil then
    FPhotos.Free;
  inherited;
end;

function TFlickrPhotoset.EditMeta(Title, Description: string): Boolean;
begin
  Result := FFlickr.PhotosetsEditMeta(FId, Title, Description);
end;

function TFlickrPhotoset.GetDescription: string;
begin
  if FDescription = FLICKR_NULL_STRING then
    FFlickr.PhotosetsGetInfo(Self);
  Result := FDescription;
end;

function TFlickrPhotoset.GetOwnerId: string;
begin
  if FOwnerId = FLICKR_NULL_STRING then
    FFlickr.PhotosetsGetInfo(Self);
  Result := FOwnerId;
end;

function TFlickrPhotoset.GetPhotos: TFlickrPhotoList;
begin
  if FPhotos = nil then
    FPhotos := FFlickr.PhotosetsGetPhotos(FId);
  Result := FPhotos;
end;

function TFlickrPhotoset.GetTitle: string;
begin
  if FTitle = FLICKR_NULL_STRING then
    FFlickr.PhotosetsGetInfo(Self);
  Result := FTitle;
end;

function TFlickrPhotoset.GetPrimaryPhotoId: string;
begin
  if FPrimaryPhotoId = FLICKR_NULL_STRING then
    FFlickr.PhotosetsGetInfo(Self);
  Result := FPrimaryPhotoId;
end;

procedure TFlickrPhotoset.Parse;
begin
  while FXmlParser.Scan do
  begin
    case FXmlParser.CurPartType of
      ptStartTag,
      ptEmptyTag:
        begin
          // read attributes
          if (FXmlParser.CurName = 'photoset') then
          begin
            if (FXmlParser.CurAttr.Count > 0) then
            begin
              SetAttr('id', FId);
              SetAttr('url', FUrl);
              SetAttr('owner', FOwnerId);
              SetAttr('primary', FPrimaryPhotoId);
            end;
          end;
          // initialize contents
          if (FXmlParser.CurName = 'title') then
            FTitle := '';
          if (FXmlParser.CurName = 'description') then
            FDescription := '';
        end;
      ptContent,
      ptCData:
        begin
          if (FXmlParser.CurName = 'title') then
            FTitle := FXmlParser.CurContent;
          if (FXmlParser.CurName = 'description') then
            FDescription := FXmlParser.CurContent;
        end;
      else
        begin
        end;
    end;
  end;
end;

function TFlickrPhotoset.EditPhotos(PhotoIds: string): Boolean;
begin
  Result := FFlickr.PhotosetsEditPhotos(FId, FPrimaryPhotoId, PhotoIds);
end;

procedure TFlickrPhotoset.ResetProperties;
begin
  FTitle := FLICKR_NULL_STRING;
  FDescription := FLICKR_NULL_STRING;
  FPrimaryPhotoId := FLICKR_NULL_STRING;
  FOwnerId := FLICKR_NULL_STRING;
  if FPhotos <> nil then FPhotos.Free;
  FPhotos := nil;
end;

{ TFlickrPhotosetList }

function TFlickrPhotosetList.Count: Integer;
begin
  Result := FPhotosetList.Count;
end;

constructor TFlickrPhotosetList.Create(Flickr: TFlickr);
begin
  FFlickr := Flickr;
  FPhotosetList := TObjectList.Create;
end;

procedure TFlickrPhotosetList.Parse;
var
  Photoset: TFlickrPhotoset;
begin
  Photoset := nil;
  while FXmlParser.Scan do
  begin
    case FXmlParser.CurPartType of
      ptStartTag,
      ptEmptyTag:
        begin
          // read attributes
          if (FXmlParser.CurName = 'photoset') then
          begin
            if (FXmlParser.CurAttr.Count > 0) then
            begin
              Photoset := TFlickrPhotoset.Create(FFlickr);
              SetAttr('id', Photoset.FId);
              SetAttr('primary', Photoset.FPrimaryPhotoId);
              FPhotosetList.Add(Photoset);
            end;
          end;
          // initialize contents
          if (FXmlParser.CurName = 'title') and (Photoset <> nil) then
            Photoset.FTitle := '';
          if (FXmlParser.CurName = 'description') and (Photoset <> nil) then
            Photoset.FDescription := '';
        end;
      ptContent,
      ptCData:
        begin
          if (FXmlParser.CurName = 'title') and (Photoset <> nil) then
            Photoset.FTitle := FXmlParser.CurContent;
          if (FXmlParser.CurName = 'description') and (Photoset <> nil) then
            Photoset.FDescription := FXmlParser.CurContent;
        end;
      else
        begin
        end;
    end;
  end;
end;

function TFlickrPhotosetList.GetPhotoset(i: Integer): TFlickrPhotoset;
begin
  Result := TFlickrPhotoset(FPhotosetList[i]);
end;

destructor TFlickrPhotosetList.Destroy;
begin
  FPhotosetList.Free;
  inherited;
end;

{ TFlickrBloglist }

procedure TFlickrBloglist.Parse;
var
  Blog: TFlickrBlog;
begin
  while FXmlParser.Scan do
  begin
    case FXmlParser.CurPartType of
      ptStartTag,
      ptEmptyTag:
        begin
          // read attributes
          if (FXmlParser.CurName = 'blog') then
          begin
            if (FXmlParser.CurAttr.Count > 0) then
            begin
              Blog := TFlickrBlog.Create(FFlickr);
              SetAttr('id', Blog.FId);
              SetAttr('name', Blog.FName);
              SetAttr('needpassword', Blog.FNeedPassword);
              SetAttr('url', Blog.FUrl);
              FBlogList.Add(Blog);
            end;
          end;
        end;
      ptContent,
      ptCData:
        begin
        end;
      else
        begin
        end;
    end;
  end;
end;

constructor TFlickrBloglist.Create(Flickr: TFlickr);
begin
  FFlickr := Flickr;
  FBlogList := TObjectList.Create;
end;

function TFlickrBloglist.Count: Integer;
begin
  Result := FBlogList.Count;
end;

function TFlickrBloglist.GetBlog(i: Integer): TFlickrBlog;
begin
  Result := TFlickrBlog(FBlogList[i]);
end;

destructor TFlickrBloglist.Destroy;
begin
  FBlogList.Free;
  inherited;
end;

{ TFlickrBlog }

function TFlickrBlog.GetNeedPassword: Boolean;
begin
  Result := FNeedPassword <> '1';
end;

constructor TFlickrBlog.Create(Flickr: TFlickr);
begin
  FFlickr := Flickr;
  FId := FLICKR_NULL_STRING;
  FName := FLICKR_NULL_STRING;
  FUrl := FLICKR_NULL_STRING;
  FNeedPassword := FLICKR_NULL_STRING;
end;

function TFlickrBlog.PostPhoto(Photo: TFlickrPhoto; Title, Description,
            BlogPassword: string): Boolean;
begin
  Result := FFlickr.BlogsPostPhoto(FId, Photo.Id, Title, Description, BlogPassword);
end;

{ TFlickrGroup }

constructor TFlickrGroup.Create(Flickr: TFlickr);
begin
  FFlickr := Flickr;
  FId := FLICKR_NULL_STRING;
  FAdmin := FLICKR_NULL_STRING;
  FPhotosCount := FLICKR_NULL_NUMBER;
  ResetProperties;
end;

function TFlickrGroup.GetPrivacy: string;
begin
  if FPrivacy = FLICKR_NULL_STRING then
    FFlickr.GroupsGetInfo(Self);
  Result := FPrivacy;
end;

function TFlickrGroup.GetChatCount: Integer;
begin
  if FChatCount = FLICKR_NULL_NUMBER then
    FFlickr.GroupsGetInfo(Self);
  Result := FChatCount;
end;

function TFlickrGroup.GetMembers: Integer;
begin
  if FMembers = FLICKR_NULL_NUMBER then
    FFlickr.GroupsGetInfo(Self);
  Result := FMembers;

end;

function TFlickrGroup.GetChatId: string;
begin
  if FChatId = FLICKR_NULL_STRING then
    FFlickr.GroupsGetInfo(Self);
  Result := FChatId;
end;

function TFlickrGroup.GetOnline: Integer;
begin
  if FOnline = FLICKR_NULL_NUMBER then
    FFlickr.GroupsGetInfo(Self);
  Result := FOnline;
end;

function TFlickrGroup.GetName: string;
begin
  if FName = FLICKR_NULL_STRING then
    FFlickr.GroupsGetInfo(Self);
  Result := FName;
end;

procedure TFlickrGroup.Parse;
begin
  while FXmlParser.Scan do
  begin
    case FXmlParser.CurPartType of
      ptStartTag,
      ptEmptyTag:
        begin
          // read attributes
          if (FXmlParser.CurName = 'group') then
          begin
            if (FXmlParser.CurAttr.Count > 0) then
            begin
              SetAttr('id', FId);
            end;
          end;
          // initialize contents
          if (FXmlParser.CurName = 'name') then
            FName := '';
          if (FXmlParser.CurName = 'members') then
            FMembers := -1;
          if (FXmlParser.CurName = 'online') then
            FOnline := -1;
          if (FXmlParser.CurName = 'privacy') then
            FPrivacy := '';
          if (FXmlParser.CurName = 'chatid') then
            FChatId := '';
          if (FXmlParser.CurName = 'chatcount') then
            FChatCount := -1;
        end;
      ptContent,
      ptCData:
        begin
          if (FXmlParser.CurName = 'name') then
            FName := FXmlParser.CurContent;
          if (FXmlParser.CurName = 'members') then
            try FMembers := StrToInt(FXmlParser.CurContent); except FMembers := -1; end;
          if (FXmlParser.CurName = 'online') then
            try FOnline := StrToInt(FXmlParser.CurContent); except FOnline := -1; end;
          if (FXmlParser.CurName = 'privacy') then
            FPrivacy := FXmlParser.CurContent;
          if (FXmlParser.CurName = 'chatid') then
            FChatId := FXmlParser.CurContent;
          if (FXmlParser.CurName = 'chatcount') then
            try FChatCount := StrToInt(FXmlParser.CurContent); except FChatCount := -1; end;
        end;
      else
        begin
        end;
    end;
  end;
end;

function TFlickrGroup.PoolPhotoAdd(Photo: TFlickrPhoto): Boolean;
begin
  Result := FFlickr.GroupsPoolsAdd(Photo.Id, FId);
end;

function TFlickrGroup.PoolPhotoRemove(Photo: TFlickrPhoto): Boolean;
begin
  Result := FFlickr.GroupsPoolsRemove(Photo.Id, FId);
end;

function TFlickrGroup.GetPhotos(PTags: string = '';
            PPerPage: Integer = -1; PPage: Integer = -1): TFlickrPhotoList;
begin
  if FPhotos = nil then
    FPhotos := FFlickr.GroupsPoolsGetPhotos(FId, PTags, PPerPage, PPage);
  Result := FPhotos;
end;

destructor TFlickrGroup.Destroy;
begin
  if FPhotos <> nil then
    FPhotos.Free;
  inherited;
end;

function TFlickrGroup.GetUrl: string;
begin
  if FUrl = FLICKR_NULL_STRING then
    FUrl := FFlickr.UrlsGetGroup(FId);
  Result := FUrl;
end;

procedure TFlickrGroup.ResetProperties;
begin
  FName := FLICKR_NULL_STRING;
  FMembers := FLICKR_NULL_NUMBER;
  FOnline := FLICKR_NULL_NUMBER;
  FPrivacy := FLICKR_NULL_STRING;
  FChatId := FLICKR_NULL_STRING;
  FChatCount := FLICKR_NULL_NUMBER;
  FUrl := FLICKR_NULL_STRING;
  if FPhotos = nil then FPhotos.Free;
  FPhotos := nil;
end;

{ TFlickrGroupList }

procedure TFlickrGroupList.Parse;
var
  Group: TFlickrGroup;
  GroupCat: TFlickrGroupCat;
begin
  while FXmlParser.Scan do
  begin
    case FXmlParser.CurPartType of
      ptStartTag,
      ptEmptyTag:
        begin
          // read attributes
          if (FXmlParser.CurName = 'group') then
          begin
            if (FXmlParser.CurAttr.Count > 0) then
            begin
              Group := TFlickrGroup.Create(FFlickr);
              SetAttr('nsid', Group.FId);
              SetAttr('id', Group.FId);
              SetAttr('name', Group.FName);
              SetAttr('members', Group.FMembers);
              SetAttr('online', Group.FOnline);
              SetAttr('admin', Group.FAdmin);
              SetAttr('privacy', Group.FPrivacy);
              SetAttr('photos', Group.FPhotosCount);
              SetAttr('chatnsid', Group.FChatId);
              SetAttr('inchat', Group.FChatCount);
              FGroupList.Add(Group);
            end;
          end;
          if (FXmlParser.CurName = 'subcat') then
          begin
            if (FXmlParser.CurAttr.Count > 0) then
            begin
              GroupCat := TFlickrGroupCat.Create;
              SetAttr('id', GroupCat.FId);
              SetAttr('name', GroupCat.FName);
              SetAttr('count', GroupCat.FGroupCount);
              FGroupCatList.Add(GroupCat);
            end;
          end;
        end;
      ptContent,
      ptCData:
        begin
        end;
      else
        begin
        end;
    end;
  end;
end;

constructor TFlickrGroupList.Create(Flickr: TFlickr);
begin
  FFlickr := Flickr;
  FGroupList := TObjectList.Create;
  FGroupCatList := TObjectList.Create;
end;

function TFlickrGroupList.GetGroupCat(i: Integer): TFlickrGroupCat;
begin
  Result := TFlickrGroupCat(FGroupCatList[i]);
end;

function TFlickrGroupList.GroupCount: Integer;
begin
  Result := FGroupList.Count;
end;

function TFlickrGroupList.GetGroup(i: Integer): TFlickrGroup;
begin
  Result := TFlickrGroup(FGroupList[i]);
end;

function TFlickrGroupList.GroupCatCount: Integer;
begin
  Result := FGroupCatList.Count;
end;

destructor TFlickrGroupList.Destroy;
begin
  FGroupList.Free;
  FGroupCatList.Free;
  inherited;
end;

end.
